import json
import datetime

from ..common import consts, error_message, json_defaul_handler
from ..common import to_bytes, to_str

class WhereCondition(object):
    def __init__(self, db):
        self.db = db
        self.queries = []
        self.params = []

    @property
    def next_literal(self):
        return self.db.value_literal(len(self.params) + 1)

    def add(self, query, param=None):
        self.queries.append(query)
        if not param is None:
            self.params.append(param)

    def add_or(self, query, param):
        self.queries.append(query)
        self.params += param

    @property
    def where_query(self):
        result = ' AND '.join(self.queries)
        if result:
            result = ' WHERE ' + result
        return result, self.params

    @property
    def or_query(self):
        result = ' OR '.join(self.queries)
        return '(%s)' % result, self.params


class AbstractDB(object):
    def __init__(self):
        self.db_type = None
        self.db_info = None
        self.DDL_ROLLBACK = False
        self.NEED_GENERATOR = False
        self.FROM = '"%s" AS %s'
        self.LIKE = 'LIKE'
        self.LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" AS %s'
        self.IS_DISTINCT_FROM = '%s IS DISTINCT FROM %s'
        self.FIELD_AS = 'AS'
        self.DESC_NULLS = None
        self.ASC_NULLS = None

    @property
    def params(self):
        return {
            'name': None,
            'dsn': False,
            'lib': [],
            'server': False,
            'database': True,
            'login': False,
            'password': False,
            'encoding': False,
            'host': False,
            'port': False,
            'ddl_rollback': self.DDL_ROLLBACK,
            'generator': self.NEED_GENERATOR,
            'import_support': True
        }
    @property
    def arg_params(self):
        return False

    def value_literal(self, index):
        return '?'

    def identifier_case(self, name):
        return name.upper()

    def convert_like(self, field_name, val, data_type):
        return 'UPPER(%s)' % field_name, val.upper()

    def next_sequence(self, table_name):
        pass

    def before_restart_sequence(self, gen_name):
        pass

    def restart_sequence(self, table_name, value):
        pass

    def process_query_params(self, params, cursor):
        result = []
        for p in params:
            if type(p) == tuple:
                value, data_type = p
            else:
                value = p
            result.append(value)
        return result

    def process_query_result(self, rows):
        return [list(row) for row in rows]

    def default_text(self, field_info):
        result = field_info.default_value
        if not result is None:
            if field_info.data_type in [consts.TEXT, consts.LONGTEXT, consts.FILE, consts.IMAGE]:
                if field_info.default_value:
                    result =  "'%s'" % result
                else:
                    result = None
            elif field_info.data_type == consts.BOOLEAN:
                if result == 'true':
                    result = '1'
                elif result == 'false':
                    result = '0'
                else:
                    result = None
            elif field_info.data_type in [consts.DATE, consts.DATETIME]:
                result = None
        return result

    def get_field_info(self, field_name, table_name, db_name=None):
        connection = self.connect(self.app.admin.task_db_info)
        try:
            fields_info = self.get_table_info(connection, table_name, db_name)
        finally:
            connection.close()
        for field in fields_info['fields']:
            if field['field_name'] == field_name:
                return field

    def before_insert(self, cursor, pk_field):
        pass

    def after_insert(self, cursor, pk_field):
        pass

    def returning(self, pk_field):
        return ''

    def insert_query(self, pk_field):
        return 'INSERT INTO "%s" (%s) VALUES (%s)'

    def db_field(self, field):
        if not field.master_field and not field.calculated:
            return True

    def insert_record(self, delta, cursor):
        if delta._deleted_flag:
            delta._deleted_flag_field.data = 0
        pk = delta._primary_key_field
        self.before_insert(cursor, pk)
        row = []
        fields = []
        values = []
        index = 0
        for field in delta.fields:
            if self.db_field(field) and not (field == pk and not pk.data):
                index += 1
                fields.append('"%s"' % field.db_field_name)
                values.append('%s' % self.value_literal(index))
                if field.data is None and not field.default_value is None:
                    field.data = field.get_default_value()
                value = (field.data, field.data_type)
                row.append(value)
        fields = ', '.join(fields)
        values = ', '.join(values)
        sql = self.insert_query(pk) % (delta.table_name, fields, values)
        row = self.process_query_params(row, cursor)
        delta.execute_query(cursor, sql, row, arg_params=self.arg_params)
        if pk:
            self.after_insert(cursor, pk)

    def check_record_version(self, delta, cursor):
        if delta.lock_active:
            if delta._record_version_field.value != delta._record_version_field.cur_value:
                raise Exception(consts.language('edit_record_modified'))
            if delta.change_log.record_status == consts.RECORD_DETAILS_MODIFIED:
                delta._record_version_field.value += 1
                delta.execute_query('UPDATE "%s" SET "%s"=COALESCE("%s", 0)+1 WHERE "%s"=%s' % \
                    (delta.table_name, delta._record_version_db_field_name,
                        delta._record_version_db_field_name, delta._primary_key_db_field_name,
                        delta._record_version_field.value))

    def update_record(self, delta, cursor):
        self.check_record_version(delta, cursor)
        row = []
        fields = []
        index = 0
        pk = delta._primary_key_field
        command = 'UPDATE "%s" SET ' % delta.table_name
        for field in delta.fields:
            prohibited, read_only = field.restrictions
            if self.db_field(field) and field != pk and not read_only:
                index += 1
                fields.append('"%s"=%s' % (field.db_field_name, self.value_literal(index)))
                if field.field_name == delta._record_version:
                    field.value += 1
                value = (field.data, field.data_type)
                if field.field_name == delta._deleted_flag:
                    value = (0, field.data_type)
                row.append(value)
        fields = ', '.join(fields)
        if delta._primary_key_field.data_type == consts.TEXT:
            id_literal = "'%s'" % delta._primary_key_field.value
        else:
            id_literal = "%s" % delta._primary_key_field.value
        where = ' WHERE "%s" = %s' % (delta._primary_key_db_field_name, id_literal)
        sql = ''.join([command, fields, where])
        row = self.process_query_params(row, cursor)
        delta.execute_query(cursor, sql, row, arg_params=self.arg_params)

    def delete_record(self, delta, cursor):
        soft_delete = delta.soft_delete
        if delta.master:
            soft_delete = delta.owner.soft_delete
        if delta._primary_key_field.data_type == consts.TEXT:
            id_literal = "'%s'" % delta._primary_key_field.value
        else:
            id_literal = "%s" % delta._primary_key_field.value
        if soft_delete:
            sql = 'UPDATE "%s" SET "%s" = 1 WHERE "%s" = %s' % \
                (delta.table_name, delta._deleted_flag_db_field_name,
                delta._primary_key_db_field_name, id_literal)
        else:
            sql = 'DELETE FROM "%s" WHERE "%s" = %s' % \
                (delta.table_name, delta._primary_key_db_field_name, id_literal)
        delta.execute_query(cursor, sql)

    def check_lookup_refs(self, delta, connection, cursor):
        if delta._lookup_refs:
            for item, fields in delta._lookup_refs.items():
                for field in fields:
                    copy = item.copy(filters=False, details=False, handlers=False)
                    field_name = field.field_name
                    #copy.set_where({field_name: delta.id.value})
                    copy.set_where({field_name: delta._primary_key_field.value})
                    copy.set_fields([field.field_name])
                    copy.open(expanded=False, limit=1, connection=connection)
                    if copy.rec_count:
                        raise Exception(consts.language('cant_delete_used_record'))

    def get_user(self, delta):
        user = None
        if delta.session:
            try:
                user = delta.session.get('user_info')['user_name']
            except:
                pass
        return user

    def init_history(self, delta):
        if delta.task.history_item and delta.keep_history and delta.change_log.record_status != consts.RECORD_DETAILS_MODIFIED:
            delta.init_history()

    def save_history(self, delta, connection, cursor):
        if delta.task.history_item and delta.keep_history and delta.change_log.record_status != consts.RECORD_DETAILS_MODIFIED:
            changes = None
            user = self.get_user(delta)
            item_id = delta.ID
            if delta.master:
                item_id = delta.prototype.ID
            if delta.change_log.record_status != consts.RECORD_DELETED:
                f_list = []
                for f in delta.fields:
                    if self.db_field(f) and f.cur_data != f.data:
                        if not delta._record_version or f.field_name != delta._record_version:
                            f_list.append([f.ID, f.data])
                if f_list:
                    changes_str = json.dumps(f_list, separators=(',',':'), default=json_defaul_handler)
                    changes = ('%s%s' % ('0', changes_str), consts.LONGTEXT)
            if changes or delta.change_log.record_status == consts.RECORD_DELETED:
                params = [item_id, delta._primary_key_field.value, delta.change_log.record_status, changes, user, datetime.datetime.now()]
                params = self.process_query_params(params, cursor)
                delta.execute_query(cursor, delta.task.history_sql, params, arg_params=self.arg_params)

    def update_deleted_detail(self, delta, detail, cursor):
        fields = [detail._primary_key]
        detail.open(fields=fields, open_empty=True)
        if detail.master_field:
            sql = 'SELECT "%s" FROM "%s" WHERE "%s" = %s AND "%s" = 0' % \
                (detail._primary_key_db_field_name, detail.table_name,
                detail._master_field_db_field_name, delta._primary_key_field.value,
                detail._deleted_flag_db_field_name)
        else:
            sql = 'SELECT "%s" FROM "%s" WHERE "%s" = %s AND "%s" = %s AND "%s" = 0' % \
                (detail._primary_key_db_field_name, detail.table_name,
                detail._master_id_db_field_name, delta.ID,
                detail._master_rec_id_db_field_name, delta._primary_key_field.value,
                detail._deleted_flag_db_field_name)
        try:
            cursor.execute(sql)
            rows = self.process_query_result(cursor.fetchall())
        except Exception as x:
            delta.log.exception(error_message(x))
            raise Exception(x)
        detail._dataset = rows

    def delete_detail_records(self, delta, connection, cursor, detail):
        if delta._primary_key_field.data_type == consts.TEXT:
            id_literal = "'%s'" % delta._primary_key_field.value
        else:
            id_literal = "%s" % delta._primary_key_field.value
        if detail._master_field:
            if delta.soft_delete:
                sql = 'UPDATE "%s" SET "%s" = 1 WHERE "%s" = %s' % \
                    (detail.table_name, detail._deleted_flag_db_field_name, \
                    detail._master_field_db_field_name, id_literal)
            else:
                sql = 'DELETE FROM "%s" WHERE "%s" = %s' % \
                    (detail.table_name, detail._master_field_db_field_name, id_literal)
        elif detail._master_id:
            if delta.soft_delete:
                sql = 'UPDATE "%s" SET "%s" = 1 WHERE "%s" = %s AND "%s" = %s' % \
                    (detail.table_name, detail._deleted_flag_db_field_name, detail._master_id_db_field_name, \
                    delta.ID, detail._master_rec_id_db_field_name, id_literal)
            else:
                sql = 'DELETE FROM "%s" WHERE "%s" = %s AND "%s" = %s' % \
                    (detail.table_name, detail._master_id_db_field_name, delta.ID, \
                    detail._master_rec_id_db_field_name, id_literal)
        elif detail._master_rec_id:
            if delta.soft_delete:
                sql = 'UPDATE "%s" SET "%s" = 1 WHERE "%s" = %s' % \
                    (detail.table_name, detail._deleted_flag_db_field_name, \
                    detail._master_rec_id_db_field_name, id_literal)
            else:
                sql = 'DELETE FROM "%s" WHERE "%s" = %s' % \
                    (detail.table_name, detail._master_rec_id_db_field_name, id_literal)
        if len(detail.details) or detail.keep_history:
            self.update_deleted_detail(delta, detail, cursor)
            if delta.task.history_item and detail.keep_history:
                for d in detail:
                    params = [detail.prototype.ID, d._primary_key_field.data,
                        consts.RECORD_DELETED, None, self.get_user(delta), datetime.datetime.now()]
                    delta.execute_query(cursor, delta.task.history_sql, \
                        self.process_query_params(params, cursor), arg_params=self.arg_params)
            if len(detail.details):
                for it in detail:
                    for d in detail.details:
                        self.delete_detail_records(detail, connection, cursor, d)
        delta.execute_query(cursor, sql)

    def process_record(self, delta, connection, cursor, safe):
        self.init_history(delta)
        if delta.master:
            if delta._master_field:
                delta._master_field_field.data = delta.master._primary_key_field.value
            elif delta._master_id:
                delta._master_id_field.data = delta.master.ID
                delta._master_rec_id_field.data = delta.master._primary_key_field.value
            elif delta._master_rec_id:
                delta._master_rec_id_field.data = delta.master._primary_key_field.value
        if delta.change_log.record_status == consts.RECORD_INSERTED:
            if safe and not delta.can_create():
                raise Exception(consts.language('cant_create') % delta.item_caption)
            self.insert_record(delta, cursor)
        elif delta.change_log.record_status == consts.RECORD_MODIFIED:
            if safe and not delta.can_edit():
                raise Exception(consts.language('cant_edit') % delta.item_caption)
            self.update_record(delta, cursor)
        elif delta.change_log.record_status == consts.RECORD_DETAILS_MODIFIED:
            self.check_record_version(delta, cursor)
        elif delta.change_log.record_status == consts.RECORD_DELETED:
            if safe and not delta.can_delete():
                raise Exception(consts.language('cant_delete') % delta.item_caption)
            self.check_lookup_refs(delta, connection, cursor)
            self.delete_record(delta, cursor)
        else:
            raise Exception('execute_delta - invalid %s record_status %s, record: %s' % \
                (delta.item_name, delta.change_log.record_status, delta._dataset[delta.rec_no]))
        self.save_history(delta, connection, cursor)

    def do_before_apply(self, delta, connection, params):
        item = delta._tree_item
        delta_params = item._get_delta_params(delta, params)
        if delta.task.on_before_apply_record:
            result = delta.task.on_before_apply_record(item, delta, delta_params, connection)
            if result == False:
                return result
        if item.on_before_apply_record:
            result = item.on_before_apply_record(item, delta, delta_params, connection)
            if result == False:
                return result

    def do_after_apply(self, delta, connection, params):
        item = delta._tree_item
        delta_params = item._get_delta_params(delta, params)
        if delta.task.on_after_apply_record:
            delta.task.on_after_apply_record(item, delta, delta_params, connection)
        if item.on_after_apply_record:
            item.on_after_apply_record(item, delta, delta_params, connection)

    def process_records(self, delta, connection, cursor, params, safe):
        for d in delta:
            res = self.do_before_apply(delta, connection, params)
            details = []
            if res != False and not d.virtual_table:
                self.process_record(d, connection, cursor, safe)
                for detail in d.details:
                    if d.change_log.record_status == consts.RECORD_DELETED:
                        self.delete_detail_records(d, connection, cursor, detail)
                    elif detail.master:
                        self.process_records(detail, connection, cursor, params, safe)
            self.do_after_apply(delta, connection, params)

    def process_changes(self, delta, connection, params=None):
        safe = delta.client_changes
        changes = []
        cursor = connection.cursor()
        self.process_records(delta, connection, cursor, params, safe)

    def table_alias(self, item):
        return '"%s"' % item.table_name

    def lookup_table_alias(self, item, field):
        if field.master_field:
            return '%s_%d' % (field.lookup_item.table_name, field.master_field.ID)
        else:
            return '%s_%d' % (field.lookup_item.table_name, field.ID)

    def lookup_table_alias1(self, item, field):
        return self.lookup_table_alias(item, field) + '_' + field.lookup_db_field

    def lookup_table_alias2(self, item, field):
        return self.lookup_table_alias1(item, field) + '_' + field.lookup_db_field1

    def field_alias(self, item, field):
        return '%s_%s' % (field.db_field_name, self.identifier_case('LOOKUP'))

    def lookup_field_sql(self, item, field):
        if field.lookup_item:
            if field.lookup_field2:
                field_sql = '%s."%s"' % (self.lookup_table_alias2(item, field), field.lookup_db_field2)
            elif field.lookup_field1:
                field_sql = '%s."%s"' % (self.lookup_table_alias1(item, field), field.lookup_db_field1)
            else:
                if field.lookup_field == field.lookup_item._primary_key:
                    field_sql = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                else:
                    field_sql = '%s."%s"' % (self.lookup_table_alias(item, field), field.lookup_db_field)
            return field_sql

    def calculated_sql(self, item, field):
        result = 'SELECT %s("%s") FROM "%s" WHERE %s.%s=%s' % \
            (field._calc_op, field._calc_field.db_field_name, field._calc_item.table_name,
            self.table_alias(item), field._calc_item._primary_key_db_field_name, field._calc_on_field.db_field_name)
        if field._calc_item._deleted_flag:
            result = '%s AND "%s"=0' % (result, field._calc_item._deleted_flag_db_field_name)
        result = '(%s) %s %s' % (result, self.FIELD_AS, self.identifier_case(field.field_name))
        return result

    def fields_clause(self, item, query, fields):
        summary = query.summary
        funcs = query.funcs
        if funcs:
            functions = {}
            for key, value in funcs.items():
                functions[key.upper()] = value
        sql = []
        for i, field in enumerate(fields):
#            if query.client_request:
#                prohibited, read_only = field.restrictions
#                if prohibited:
#                    print(field.field_name)
#                    continue
            if i == 0 and summary:
                sql.append(self.identifier_case('count(*)'))
            elif field.master_field:
                pass
            elif field.calculated:
                pass
            else:
                field_sql = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                func = None
                if funcs:
                    func = functions.get(field.field_name.upper())
                    if func:
                        field_sql = '%s(%s) %s "%s"' % (func.upper(), field_sql, self.FIELD_AS, field.db_field_name)
                sql.append(field_sql)
        for i, field in enumerate(fields):
            if field.calculated:
                if query.expanded:
                    sql.append(self.calculated_sql(item, field))
        if query.expanded:
            for i, field in enumerate(fields):
                if i == 0 and summary:
                    continue
                field_sql = self.lookup_field_sql(item, field)
                field_alias = self.field_alias(item, field)
                if field_sql:
                    if funcs:
                        func = functions.get(field.field_name.upper())
                    if func:
                        field_sql = '%s(%s) %s "%s"' % (func.upper(), field_sql, self.FIELD_AS, field_alias)
                    else:
                        field_sql = '%s %s %s' % (field_sql, self.FIELD_AS, field_alias)
                    sql.append(field_sql)
        sql = ', '.join(sql)
        return sql

    def add_join(self, result, joins, item, field):
        alias = self.lookup_table_alias(item, field)
        cur_field = field
        if field.master_field:
            cur_field = field.master_field
        if not joins.get(alias):
            primary_key_field_name = field.lookup_item._primary_key_db_field_name
            result.append('%s ON %s."%s" = %s."%s"' % (
                self.LEFT_OUTER_JOIN % (field.lookup_item.table_name, self.lookup_table_alias(item, field)),
                self.table_alias(item),
                cur_field.db_field_name,
                self.lookup_table_alias(item, field),
                primary_key_field_name
            ))
            joins[alias] = True

    def add_join1(self, result, joins, item, field):
        alias = self.lookup_table_alias1(item, field)
        if not joins.get(alias):
            primary_key_field_name = field.lookup_item1._primary_key_db_field_name
            result.append('%s ON %s."%s" = %s."%s"' % (
                self.LEFT_OUTER_JOIN % (field.lookup_item1.table_name, self.lookup_table_alias1(item, field)),
                self.lookup_table_alias(item, field),
                field.lookup_db_field,
                self.lookup_table_alias1(item, field),
                primary_key_field_name
            ))
            joins[alias] = True

    def add_join2(self, result, joins, item, field):
        alias = self.lookup_table_alias2(item, field)
        if not joins.get(alias):
            primary_key_field_name = field.lookup_item2._primary_key_db_field_name
            result.append('%s ON %s."%s" = %s."%s"' % (
                self.LEFT_OUTER_JOIN % (field.lookup_item2.table_name, self.lookup_table_alias2(item, field)),
                self.lookup_table_alias1(item, field),
                field.lookup_db_field1,
                self.lookup_table_alias2(item, field),
                primary_key_field_name
            ))
            joins[alias] = True

    def from_clause(self, item, query, fields):
        result = []
        result.append(self.FROM % (item.table_name, self.table_alias(item)))
        fields = list(fields)
        filters = query.filters
        if filters:
            for f in filters:
                if type(f[0]) != list:
                    field_name, filter_type, value = f
                    if not value is None:
                        field = item._field_by_name(field_name)
                        if not field in fields:
                            fields.append(field)
        if query.expanded:
            joins = {}
            for field in fields:
                if field.lookup_item and field.data_type != consts.KEYS:
                    self.add_join(result, joins, item, field)
                if field.lookup_item1:
                    self.add_join1(result, joins, item, field)
                if field.lookup_item2:
                    self.add_join2(result, joins, item, field)
            filters = query.filters
        return ' '.join(result)

    def get_filter_sign(self, item, filter_type, value):
        result = consts.FILTER_SIGN[filter_type]
        if filter_type == consts.FILTER_ISNULL:
            if value:
                result = 'IS NULL'
            else:
                result = 'IS NOT NULL'
        if (result == 'LIKE'):
            result = self.LIKE
        return result

    def convert_field_value(self, field, value):
        data_type = field.data_type
        if isinstance(value, str):
            value = to_str(value)
        if data_type == consts.DATE:
            if isinstance(value, str):
                return datetime.datetime.strptime(value, '%Y-%m-%d')
        if data_type == consts.DATETIME:
            if isinstance(value, str):
                return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        if data_type == consts.BOOLEAN:
            if value:
                return '1'
            else:
                return '0'
        else:
            return value

    def escape_search(self, value, esc_char):
        result = ''
        found = False
        for ch in value:
            if ch == "'":
                ch = ch + ch
            elif ch in ['_', '%']:
                ch = esc_char + ch
                found = True
            result += ch
        return result, found

    def get_condition(self, item, conditions, field, filter_type, value):
        esc_char = '/'
        cond_field_name = '%s."%s"' % (self.table_alias(item), field.db_field_name)
        if filter_type > consts.FILTER_CONTAINS:
            if field.lookup_item:
                if field.lookup_item1:
                    cond_field_name = '%s."%s"' % (self.lookup_table_alias1(item, field), field.lookup_db_field1)
                else:
                    if field.data_type == consts.KEYS:
                        cond_field_name = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                    else:
                        cond_field_name = '%s."%s"' % (self.lookup_table_alias(item, field), field.lookup_db_field)
        if filter_type > consts.FILTER_CONTAINS_ALL:
            filter_type -= consts.FILTER_CONTAINS_ALL
        sql_literal = conditions.next_literal
        filter_sign = self.get_filter_sign(item, filter_type, value)
        if filter_type in (consts.FILTER_IN, consts.FILTER_NOT_IN):
            values = [to_str(v) for v in value if v is not None]
            sql_literal = '(%s)' % ', '.join(values)
            value = None
        elif filter_type == consts.FILTER_RANGE:
            param1 = self.convert_field_value(field, value[0])
            conditions.params.append(param1)
            sql_literal2 = conditions.next_literal
            param2 = self.convert_field_value(field, value[1])
            sql_literal = ' %s AND %s ' % (sql_literal, sql_literal2)
            value = param2
        elif filter_type == consts.FILTER_ISNULL:
            sql_literal = ''
            value = None
        elif filter_type in [consts.FILTER_CONTAINS, consts.FILTER_STARTWITH, consts.FILTER_ENDWITH]:
                value = self.convert_field_value(field, value)
                value, esc_found = self.escape_search(value, esc_char)
                if field.lookup_item:
                    if field.lookup_item1:
                        cond_field_name = '%s."%s"' % (self.lookup_table_alias1(item, field), field.lookup_db_field1)
                    else:
                        if field.data_type == consts.KEYS:
                            cond_field_name = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                        else:
                            cond_field_name = '%s."%s"' % (self.lookup_table_alias(item, field), field.lookup_db_field)
                if filter_type == consts.FILTER_CONTAINS:
                    value = '%' + value + '%'
                elif filter_type == consts.FILTER_STARTWITH:
                    value = value + '%'
                elif filter_type == consts.FILTER_ENDWITH:
                    value = '%' + value
                cond_field_name, value = self.convert_like(cond_field_name, value, field.data_type)
                if esc_found:
                    value = '' + value + "' ESCAPE '" + esc_char
                value = '%s' % value
        else:
            value = self.convert_field_value(field, value)
        sql = '%s %s %s' % (cond_field_name, filter_sign, sql_literal)
        if field.data_type == consts.BOOLEAN and value == 0 and filter_sign == '=':
            value = '1'
            sql = self.IS_DISTINCT_FROM % (cond_field_name, sql_literal)
        if filter_sign == '<>':
            sql = self.IS_DISTINCT_FROM % (cond_field_name, sql_literal)
        return sql, value

    def add_master_conditions(self, item, conditions, query):
        if query.master_field:
            conditions.add('%s."%s" = %s' % \
                (self.table_alias(item), item._master_field_db_field_name,
                conditions.next_literal), query.master_field)
        elif query.master_id:
            conditions.add('%s."%s" = %s' % \
                (self.table_alias(item), item._master_id_db_field_name,
                conditions.next_literal), query.master_id)
            conditions.add('%s."%s" = %s' % \
                (self.table_alias(item), item._master_rec_id_db_field_name,
                conditions.next_literal), query.master_rec_id)
        elif query.master_rec_id:
            conditions.add('%s."%s" = %s' % \
                (self.table_alias(item), item._master_rec_id_db_field_name,
                conditions.next_literal), query.master_rec_id)

    def where_clause(self, item, query, or_clause=False):
        conditions = WhereCondition(self)
        if or_clause:
            filters = query
        else:
            filters = query.filters
            if item.master:
                self.add_master_conditions(item, conditions, query)
        deleted_in_filters = False
        if filters:
            for f in filters:
                if type(f[0]) == list:
                    query, param = self.where_clause(item, f, True)
                    conditions.add_or(query, param)
                else:
                    field_name, filter_type, value = f
                    if not value is None:
                        field = item._field_by_name(field_name)
                        if field_name == item._deleted_flag:
                            deleted_in_filters = True
                        if filter_type == consts.FILTER_CONTAINS_ALL:
                            values = value.split()
                            for val in values:
                                query, param = self.get_condition(item, conditions, field, consts.FILTER_CONTAINS, val)
                                conditions.add(query, param)
                        elif filter_type in [consts.FILTER_IN, consts.FILTER_NOT_IN] and \
                            type(value) in [tuple, list] and len(value) == 0:
                            conditions.add('%s."%s" IN (NULL)' % (self.table_alias(item), item._primary_key_db_field_name))
                        else:
                            query, param = self.get_condition(item, conditions, field, filter_type, value)
                            conditions.add(query, param)
        if or_clause:
            return conditions.or_query
        else:
            if not deleted_in_filters and item._deleted_flag:
                conditions.add('%s."%s" = 0' % (self.table_alias(item), item._deleted_flag_db_field_name))
            return conditions.where_query

    def group_clause(self, item, query, fields):
        group_fields = query.group_by
        funcs = query.funcs
        if funcs:
            functions = {}
            for key, value in funcs.items():
                functions[key.upper()] = value
        result = ''
        if group_fields:
            for field_name in group_fields:
                field = item._field_by_name(field_name)
                use_lookup_field = (query.expanded and field.lookup_item and field.data_type != consts.KEYS)
                lookup_sql = self.lookup_field_sql(item, field)
                table_alias = self.table_alias(item)
                if use_lookup_field:
                    func = functions.get(field.field_name.upper())
                    if func:
                            result += '%s."%s", ' % (self.table_alias(item), field.db_field_name)
                    elif field.master_field:
                            result += '%s, ' % lookup_sql
                    elif field.lookup_field:
                        result += '%s, %s."%s", ' % (lookup_sql, table_alias, field.db_field_name)
                    else:
                            result += '%s, ' % lookup_sql
                else:
                    result += '%s."%s", ' % (table_alias, field.db_field_name)
            if result:
                result = result[:-2]
                result = ' GROUP BY ' + result
            return result
        else:
            return ''

    def order_clause(self, item, query):
        limit = query.limit
        if limit and not query.order and item._primary_key:
            query.order = [[item._primary_key, False]]
        if query.funcs and not query.group_by:
            return ''
        funcs = query.funcs
        functions = {}
        if funcs:
            for key, value in funcs.items():
                functions[key.upper()] = value
        order_list = query.order
        orders = []
        for order in order_list:
            field = item._field_by_name(order[0])
            if field:
                func = functions.get(field.field_name.upper())
                if not query.expanded and field.lookup_item1:
                   orders = []
                   break
                if query.expanded and field.lookup_item:
                    if field.data_type == consts.KEYS:
                        ord_str = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                    else:
                        if func:
                            ord_str = self.field_alias(item, field)
                        else:
                            ord_str = self.lookup_field_sql(item, field)
                elif field.calculated:
                    ord_str = self.identifier_case(field.field_name)
                else:
                    if func:
                        if self.db_type == consts.MSSQL and limit:
                            ord_str = '%s(%s."%s")' %  (func, self.table_alias(item), field.db_field_name)
                        else:
                            ord_str = '"%s"' % field.db_field_name
                    else:
                        ord_str = '%s."%s"' % (self.table_alias(item), field.db_field_name)
                if order[1]:
                    ord_str += ' DESC'
                    if self.DESC_NULLS:
                        ord_str += ' %s' % self.DESC_NULLS
                elif self.ASC_NULLS:
                    ord_str += ' ASC %s' % self.ASC_NULLS
                orders.append(ord_str)
        if orders:
             result = ' ORDER BY %s' % ', '.join(orders)
        else:
            result = ''
        return result

    def split_query(self, query):
        MAX_IN_LIST = 1000
        filters = query.filters
        filter_index = -1
        max_list = 0
        if filters:
            for i, f in enumerate(filters):
                if type(f[0]) != list: # not or filter
                    field_name, filter_type, value = f
                    if filter_type in [consts.FILTER_IN, consts.FILTER_NOT_IN]:
                        length = len(value)
                        if length > MAX_IN_LIST and length > max_list:
                            max_list = length
                            filter_index = i
        if filter_index != -1:
            lists = []
            value_list = filters[filter_index][2]
            while True:
                values = value_list[0:MAX_IN_LIST]
                if values:
                    lists.append(values)
                value_list = value_list[MAX_IN_LIST:]
                if not value_list:
                    break;
            return filter_index, lists

    def get_select_queries(self, item, query):
        result = []
        filter_in_info = self.split_query(query)
        if filter_in_info:
            filter_index, lists = filter_in_info
            for lst in lists:
                query.limit = None
                query.offset = None
                query.filters[filter_index][2] = lst
                result.append(self.get_select_query(item, query))
        else:
            result.append(self.get_select_query(item, query))
        return result

    def get_select_statement(self, item, query): # depricated
        return self.get_select_query(item, query)

    def get_select_query(self, item, query):
        try:
            field_list = query.fields
            if len(field_list):
                fields = [item._field_by_name(field_name) for field_name in field_list]
            else:
                fields = item._fields
            fields_clause = self.fields_clause(item, query, fields)
            from_clause = self.from_clause(item, query, fields)
            where_clause, params = self.where_clause(item, query)
            group_clause = self.group_clause(item, query, fields)
            order_clause = self.order_clause(item, query)
            sql = self.get_select(query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields)
            return sql, params
        except Exception as e:
            item.log.exception(error_message(e))
            raise

    def get_record_count_query(self, item, query):
        fields = []
        filters = query.filters
        if filters:
            for (field_name, filter_type, value) in filters:
                fields.append(item._field_by_name(field_name))
        where_sql, params = self.where_clause(item, query)
        sql = 'SELECT COUNT(*) FROM %s %s' % (self.from_clause(item, query, fields),
            where_sql)
        return sql, params

    def empty_table_query(self, item):
        return 'DELETE FROM %s' % item.table_name




import sqlite3

from ..common import consts
from .db import AbstractDB

class SQLiteDB(AbstractDB):
    def __init__(self):
        AbstractDB.__init__(self)
        self.db_type = consts.SQLITE
        self.IS_DISTINCT_FROM = 'NOT %s IS %s'
        self.FIELD_TYPES = {
            consts.INTEGER: 'INTEGER',
            consts.TEXT: 'TEXT',
            consts.FLOAT: 'REAL',
            consts.CURRENCY: 'REAL',
            consts.DATE: 'TEXT',
            consts.DATETIME: 'TEXT',
            consts.BOOLEAN: 'INTEGER',
            consts.LONGTEXT: 'TEXT',
            consts.KEYS: 'TEXT',
            consts.FILE: 'TEXT',
            consts.IMAGE: 'TEXT'
        }

    def get_params(self, lib):
        params = self.params
        params['name'] = 'SQLITE'
        return params

    def sqlite_upper(self, value):
        try:
            return value.upper()
        except:
            return value

    def connect(self, db_info):
        if not db_info.database:
            raise Exception('Must supply database name')
        connection = sqlite3.connect(db_info.database)
        connection.create_function("UPPER", 1, self.sqlite_upper)
        cursor = connection.cursor()
        cursor.execute("PRAGMA foreign_keys = ON")
        return connection

    def get_select(self, query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
        start = fields_clause
        end = ''.join([from_clause, where_clause, group_clause, order_clause])
        offset = query.offset
        limit = query.limit
        result = 'SELECT %s FROM %s' % (start, end)
        if limit:
            result += ' LIMIT %d, %d' % (offset, limit)
        return result

    def create_table(self, table_name, fields, gen_name=None, foreign_fields=None):
        primary_key = ''
        sql = 'CREATE TABLE "%s"\n(\n' % table_name
        lines = []
        for field in fields:
            default_text = self.default_text(field)
            line = '"%s" %s' % (field.field_name, self.FIELD_TYPES[field.data_type])
            if field.primary_key:
                primary_key = field.field_name
                line += ' PRIMARY KEY'
            if not default_text is None:
                line += ' DEFAULT %s' % default_text
            lines.append(line)
        if foreign_fields:
            for field in foreign_fields:
                lines.append('FOREIGN KEY(%s) REFERENCES %s(%s)\n' % \
                    (field['key'], field['ref'], field['primary_key']))
        sql += ',\n'.join(lines)
        sql += '\n)'
        return sql

    def add_field(self, table_name, field):
        default_text = self.default_text(field)
        result = 'ALTER TABLE "%s" ADD COLUMN "%s" %s' % \
            (table_name, field.field_name, self.FIELD_TYPES[field.data_type])
        if not default_text is None:
            result += ' DEFAULT %s' % default_text
        return result

    def del_field(self, table_name, field):
        return ''

    def change_field(self, table_name, old_field, new_field):
        return ''

    def drop_table(self, table_name, gen_name):
        return 'DROP TABLE IF EXISTS "%s"' % table_name

    def set_foreign_keys(self, value):
        if value:
            return 'PRAGMA foreign_keys=on'
        else:
            return 'PRAGMA foreign_keys=off'

    def create_index(self, index_name, table_name, unique, fields, desc):
        return 'CREATE %s INDEX "%s" ON "%s" (%s)' % (unique, index_name, table_name, fields)

    def drop_index(self, table_name, index_name):
        return 'DROP INDEX IF EXISTS "%s"' % index_name

    def after_insert(self, cursor, pk_field):
        if pk_field and not pk_field.data:
            pk_field.data = cursor.lastrowid

    def get_table_names(self, connection):
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM sqlite_master WHERE type='table'")
        result = cursor.fetchall()
        return [r[1] for r in result]

    def get_table_info(self, connection, table_name, db_name):
        cursor = connection.cursor()
        cursor.execute("PRAGMA table_info('%s')" % table_name)
        result = cursor.fetchall()
        fields = []
        for r in result:
            fields.append({
                'field_name': r[1],
                'data_type': r[2],
                'size': 0,
                'default_value': r[4],
                'pk': r[5]==1
            })
        return {'fields': fields, 'field_types': self.FIELD_TYPES}

db = SQLiteDB()





import sys
import sys
import os
import json
import uuid
import traceback
import datetime, time
from types import MethodType
import mimetypes
import logging
import jam

from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, NotFound, Forbidden
from werkzeug.middleware.shared_data import SharedDataMiddleware
from werkzeug.wsgi import get_path_info
from werkzeug.local import Local, LocalManager
from werkzeug.http import parse_date, http_date
from werkzeug.utils import cached_property, secure_filename, redirect
from .secure_cookie.cookie import SecureCookie

from .common import consts, error_message, file_read, file_write
from .common import consts, ProjectError, ProjectNotCompleted
from .common import json_defaul_handler, compressBuf
from .common import validate_image, valid_uploaded_file
from .common import to_str
from .admin.admin import create_admin, login_user, get_privileges, get_field_restrictions
from .admin.admin import user_valid_ip, user_valid_uuid
from .admin.builder import update_events_code
from .admin.import_metadata import import_metadata
from .tree import AbortException
from .admin.task import create_task, reload_task

class JamSecureCookie(SecureCookie):
    serialization_method = json

    def save_cookie(
        self,
        response,
        key="session",
        expires=None,
        session_expires=None,
        max_age=None,
        path="/",
        domain=None,
        secure=None,
        httponly=False,
        force=False,
        samesite=None
    ):
        if force or self.should_save:
            data = self.serialize(session_expires or expires)
            response.set_cookie(
                key,
                data,
                expires=expires,
                max_age=max_age,
                path=path,
                domain=domain,
                secure=secure,
                httponly=httponly,
                samesite=samesite
            )


class JamRequest(Request):

    @cached_property
    def session_key(self):
        return '%s_session_%s' % (self.task.item_name, self.environ['SERVER_PORT'])

    @cached_property
    def client_cookie(self):
        return JamSecureCookie.load_cookie(self, key=self.session_key, secret_key=self.task.app.admin.secret_key)

    def save_client_cookie(self, response, task):
        if task:
            cookie = self.client_cookie
            expires = None
            if consts.SAFE_MODE and task.timeout:
                expires = time.time() + task.timeout
                cookie.modified = True
            cookie.save_cookie(response, key=self.session_key, session_expires=expires, httponly=True, samesite='Lax')

def create_application(from_file=None, load_task=False, testing=False):
    if from_file:
        if os.path.isfile(from_file):
            work_dir = os.path.dirname(from_file)
        else:
            work_dir = from_file
    else:
        work_dir = os.getcwd()
    work_dir = os.path.realpath(work_dir)
    os.chdir(work_dir)
    static_files = {
        '/static':  os.path.join(work_dir, 'static')
    }

    jam.context = Local()
    local_manager = LocalManager([jam.context])

    application = App(work_dir, load_task)
    if not testing:
        application = SharedDataMiddleware(application, static_files)
        application = local_manager.make_middleware(application)
    return application

class JamLogger(object):
    def __init__(self, app):
        self.app = app

    @cached_property
    def log(self):
        logger = logging.getLogger('jam_app')
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        return logger

    def debug(self, msg, *args, **kwargs):
        self.log.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log.error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self.log.exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log.critical(msg, *args, **kwargs)

class App(object):
    def __init__(self, work_dir, load_task):
        mimetypes.add_type('text/cache-manifest', '.appcache')
        self.started = datetime.datetime.now()
        self.work_dir = to_str(work_dir)
        self.state = consts.PROJECT_NONE
        self.__task = None
        self.privileges = None
        self.field_restrictions = None
        self.__is_locked = 0
        self.__loading = False
        self._busy = 0
        self.pid = os.getpid()
        self.jam_dir = to_str(os.path.realpath(os.path.dirname(jam.__file__)))
        self.jam_version = jam.version()
        self.application_files = {
            '/': self.work_dir,
            '/jam/': self.jam_dir
        }
        self.fileserver = SharedDataMiddleware(None, self.application_files, cache_timeout=1)
        consts.app = self
        self.log = JamLogger(self)
        create_admin(self)
        self.build_id_prefix = '$buildID'
        self.save_build_id()
        if load_task:
            with self.admin.lock('$creating_task'):
                self.__task = create_task(self)
        self.check_migration()

    def create_task(self):
        result = self.__task
        if not result and not self.__loading:
            self.__loading = True
            try:
                result = create_task(self)
                update_events_code(self.admin, loading=True)
                consts.CLIENT_MODIFIED = False
                consts.SERVER_MODIFIED = False
                consts.write_settings()
                self.state = consts.RESPONSE
                result.__task_locked = True
            except ProjectNotCompleted:
                raise
            except:
                traceback.print_exc()
                raise ProjectError()
            finally:
                self.__loading = False
        else:
            time.sleep(1)
            result = self.__task
            if not result:
                result = self.create_task()
        return result

    @property
    def task(self):
        if not self.__task:
            self.__task = self.create_task()
        return self.__task

    def task_locked(self):
        return self.__is_locked > 0

    @property
    def __task_locked(self):
        pass

    @__task_locked.setter
    def __task_locked(self, value):
        if value:
            self.__is_locked += 1
        else:
            self.__is_locked -= 1

    @cached_property
    def production(self):
        return os.path.exists(os.path.join(self.work_dir, 'builder.html'))

    def __call__(self, environ, start_response):
        jam.context.environ = environ
        jam.context.session = None
        request = JamRequest(environ)
        if consts.MAX_CONTENT_LENGTH > 0:
            request.max_content_length = 1024 * 1024 * consts.MAX_CONTENT_LENGTH
        try:
            if not self.under_maintenance:
                if self.__task and self.task.on_request:
                    if not self.production:
                        self.check_project_modified()
                    response = self.task.on_request(self.task, request)
                    if response:
                        return response(environ, start_response)
                parts = request.path.strip('/').split('/')
                prefix = parts[0]
                suffix = parts[len(parts) - 1]
                if prefix == 'api':
                    return self.on_api(request)(environ, start_response)
                elif prefix == 'upload':
                    return self.on_upload(request)(environ, start_response)
                elif prefix == 'jam':
                    return self.on_jam_file(request, environ, suffix)(environ, start_response)
                elif prefix in ['js', 'css']:
                    return self.on_project_file(request, environ, suffix)(environ, start_response)
                elif prefix in ['builder.html', 'builder_login.html']:
                    return self.on_builder(request, prefix)(environ, start_response)
                elif prefix == 'ext':
                    return self.on_ext(request)(environ, start_response)
                elif prefix in ['static', 'favicon.ico', 'dummy.html']:
                    return Response('')(environ, start_response)
                elif not prefix or prefix in ['index.html', 'login.html']:
                    return self.on_index(request, prefix)(environ, start_response)
                elif prefix == 'logout':
                    if self.task.on_logout:
                        response = self.task.on_logout(self.task, request)
                        if response:
                            return response
                    response = self.logout(request)
                    return response(environ, start_response)
                else:
                    print(request.path)
                    raise NotFound()
        except ProjectNotCompleted as e:
            self.log.exception(error_message(e))
            return self.show_information(consts.lang['no_project'])(environ, start_response)
        except ProjectError as e:
            self.log.exception(error_message(e))
            return self.show_error(consts.lang['project_error'])(environ, start_response)
        except HTTPException as e:
            self.log.exception(error_message(e))
            return self.show_error(error_message(e))(environ, start_response)
            # ~ return e

    def serve_page(self, file_name, dic=None):
        path = os.path.join(self.work_dir, file_name)
        if os.path.exists(path):
            page = file_read(path)
            if dic:
                page = page % dic
            return Response(page, mimetype="text/html")
        else:
            raise NotFound()

    def show_information(self, message):
        return self.show_error(message, 'alert-info', 'Information')

    def show_error(self, message, error_class=None, error_type=None):
        if not error_class:
            error_class = 'alert-error'
        if not error_type:
            error_type = 'Error'
        path = os.path.join(self.jam_dir, 'html', 'error.html')
        return self.serve_page(path, {
            'error_class': error_class,
            'error_type': error_type,
            'message': message
        })

    def on_index(self, request, file_name):
        if file_name == 'login.html':
            login_params = {
                'title': self.task.item_name,
                'error': '',
                'form_title': consts.lang['log_in'],
                'login_text': consts.lang['login'],
                'password_text': consts.lang['password'],
                'login': '',
                'password': ''
            }
            login_path = os.path.join(self.work_dir, 'login.html')
            if not os.path.exists(login_path):
                login_path = os.path.join(self.jam_dir, 'html', 'login.html')
            if request.method == 'POST':
                response = self.login(request, self.task, request.form)
                if response:
                    return response
                else:
                    form = request.form.to_dict()
                    login_params['error'] = 'error-modal-border'
                    login_params['login'] = form['login']
                    login_params['password'] = form['password']
                    return self.serve_page(login_path, login_params)
            else:
                return self.serve_page(login_path, login_params)
        else:
            if self.check_session(request, self.task):
                file_name = 'index.html'
                request.environ['PATH_INFO'] = '/%s' % file_name
                path = os.path.join(self.work_dir, file_name)
                self.check_project_modified()
                return self.fileserver
            else:
                return redirect('/login.html')

    def on_builder(self, request, file_name):
        if file_name == 'builder.html':
            if os.path.exists(os.path.join(self.work_dir, file_name)):
                request.environ['PATH_INFO'] = '/%s' % file_name
                return self.fileserver
            if self.check_session(request, self.admin):
                request.environ['PATH_INFO'] = '/jam/html/%s' % file_name
                return self.fileserver
            else:
                return redirect('/builder_login.html')
        elif file_name == 'builder_login.html':
            login_params = {
                'title': 'Jam.py Application Builder',
                'error': '',
                'form_title': consts.lang['log_in'] + ' - Application Builder',
                'login_text': consts.lang['login'],
                'password_text': consts.lang['password'],
                'login': '',
                'password': ''
            }
            login_path = os.path.join(self.jam_dir, 'html', 'login.html')
            if request.method == 'POST':
                response = self.login(request, self.admin, request.form)
                if response:
                    return response
                else:
                    form = request.form.to_dict()
                    login_params['error'] = 'error-modal-border'
                    login_params['login'] = form['login']
                    login_params['password'] = form['password']
                    return self.serve_page(login_path, login_params)
            else:
                return self.serve_page(login_path, login_params)

    def serve_prog_file(self, request, environ, file_name):
        base, ext = os.path.splitext(file_name)
        if consts.COMPRESSED_JS and ext and ext in ['.js', '.css'] and file_name != 'project.css':
            min_file_name = base + '.min' + ext
            environ['PATH_INFO'] = environ['PATH_INFO'].replace(file_name, min_file_name)
        return self.fileserver

    def on_jam_file(self, request, environ, file_name):
        return self.serve_prog_file(request, environ, file_name)

    def on_project_file(self, request, environ, file_name):
        if self.check_session(request, self.task):
            return self.serve_prog_file(request, environ, file_name)
        else:
            raise Forbidden()

    def create_post_response(self, request, result):
        response = Response()
        response.headers['Content-Type'] = 'application/json'
        try:
            buff = json.dumps(result, default=json_defaul_handler)
        except Exception as e:
            self.log.exception('wsgi.py create_post_response error %s' % error_message(e))
            self.log.debug(result)
            raise
        accept_encoding = request.headers.get('Accept-Encoding', '')
        if 'gzip' not in accept_encoding.lower():
            response.set_data(buff)
            return response
        buff = compressBuf(buff)
        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Vary'] = 'Accept-Encoding'
        response.headers['Content-Length'] = len(buff)
        response.set_data(buff)
        return response

    def get_client_address(self, request):
        try:
            return request.environ['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()
        except KeyError:
            return request.environ['REMOTE_ADDR']

    def create_session(self, request, task, user_info=None, session_uuid=None):
        if not user_info:
            user_info = {
                'user_id': None,
                'role_id': None,
                'role_name': '',
                'user_name': '',
                'admin': False
            }
        cookie = request.client_cookie
        session = {}
        session['ip'] = self.get_client_address(request)
        session['uuid'] = session_uuid
        session['user_info'] = user_info
        cookie['info'] = session
        cookie.modified = True
        return cookie

    def connect(self, request, task):
        if self.check_session(request, task):
            return consts.PROJECT_LOGGED

    def valid_session(self, task, session, request):
        if consts.SAFE_MODE:
            user_info = session['user_info']
            if not (user_info and user_info.get('user_id')):
                return False
            if not self.admin.ignore_change_ip and task != self.admin:
                ip = self.get_client_address(request)
                if not user_valid_ip(self.admin, user_info['user_id'], ip):
                    return False
            if not self.admin.ignore_change_uuid and task != self.admin:
                if not user_valid_uuid(self.admin, user_info['user_id'], session['uuid']):
                    return False
        return True

    def check_session(self, request, task):
        request.task = task
        c = request.client_cookie
        if not c.get('info') and not consts.SAFE_MODE:
            c = self.create_session(request, task)
        session = c.get('info')
        if session:
            if not self.valid_session(task, session, request):
                self.logout(request, task)
                return False
            jam.context.session = session
            return True

    def default_login(self, task, login, password, ip, session_uuid):
        return login_user(self.admin, login, password, self.admin == task, ip, session_uuid)

    def login(self, request, task, form_data):
        time.sleep(0.3)
        request.task = task
        ip = None
        session_uuid = None
        ip = self.get_client_address(request)
        session_uuid = str(uuid.uuid4())
        if self.admin == task or task.on_login is None:
            user_info = self.default_login(task, form_data['login'], form_data['password'], ip, session_uuid)
        elif task.on_login:
            try:
                user_info = task.on_login(task, form_data, {'ip': ip, 'session_uuid': session_uuid})
            except:
                user_info = None
                traceback.print_exc()
        if user_info:
            self.create_session(request, task, user_info, session_uuid)
            if self.admin == task:
                response = redirect('/builder.html')
            else:
                response = redirect('/')
            request.save_client_cookie(response, task)
            return response


    def logout(self, request, task=None):
        if task is None:
            task = self.task
        request.task = task
        response = None
        if self.admin == task:
            del request.client_cookie['info']
        else:
            response = redirect('/')
            del request.client_cookie['info']
            request.save_client_cookie(response, self.task)
        jam.context.session = None
        return response

    def get_role_privileges(self, role_id):
        if consts.SAFE_MODE:
            if self.privileges is None:
                self.privileges = get_privileges(self.admin)
            return self.privileges[role_id]

    def get_role_field_restrictions(self, role_id):
        if consts.SAFE_MODE:
            if self.field_restrictions is None:
                self.field_restrictions = get_field_restrictions(self.admin)
            return self.field_restrictions[role_id]
        else:
            return {}

    def init_client(self, task):
        session = jam.context.session
        priv = None
        user_info = {}
        if session:
            user_info = session['user_info']
            role_id = user_info.get('role_id')
            if role_id:
                priv = self.get_role_privileges(role_id)
        templates = ''
        templates_path = os.path.join(self.work_dir, 'templates.html')
        if not os.path.exists(templates_path):
            file_write(templates_path, '')
        if task != self.admin:
            templates = file_read(templates_path)
        result = {
            'task': task.get_info(),
            'templates': templates,
            'settings': consts.settings,
            'locale': consts.locale,
            'language': consts.lang,
            'user_info': user_info,
            'privileges': priv
        }
        return result, ''

    def check_migration(self):
        path = os.path.join(self.work_dir, 'migration')
        files = []
        if os.path.exists(path):
            for file_name in os.listdir(path):
                files.append(os.path.join(self.work_dir, 'migration', file_name))
        files_len = len(files)
        if files_len:
            if files_len == 1:
                self.import_md(files[0], False)
            else:
                self.log.error('More than one file in migration folder')

    def import_metadata(self, task, task_id, file_name, from_client): #for compatibility with previous versions
        self.import_md(file_name, from_client)

    def import_md(self, file_name, from_client):
        if not self.under_maintenance:
            with self.admin.lock('$metadata_import'):
                success = False
                consts.MAINTENANCE = True
                consts.PARAMS_VERSION += 1
                consts.write_settings()
                self.save_build_id()
                self.__task_locked = False
                try:
                    result = import_metadata(self.admin, file_name, from_client)
                    success, error, message = result
                    if success and self.task:
                        reload_task(self)
                        update_events_code(self.admin)
                        self.privileges = None
                except:
                    self.log.exception('wsgi.py create_post_response error')
                finally:
                    self.__task_locked = True
                    consts.MAINTENANCE = False
                    consts.PARAMS_VERSION += 1
                    if success:
                        consts.BUILD_VERSION += 1
                        consts.MODIFICATION += 1
                    consts.write_settings()
                    self.save_build_id()
                return result

    @property
    def under_maintenance(self):
        return consts.MAINTENANCE

    @property
    def client_modified(self):
        return consts.CLIENT_MODIFIED

    @client_modified.setter
    def client_modified(self, value):
        consts.CLIENT_MODIFIED = value
        consts.MODIFICATION += 1
        consts.PARAMS_VERSION += 1
        consts.write_settings()
        self.save_build_id()

    @property
    def server_modified(self):
        return consts.SERVER_MODIFIED

    @server_modified.setter
    def server_modified(self, value):
        consts.SERVER_MODIFIED = value
        consts.MODIFICATION += 1
        consts.PARAMS_VERSION += 1
        consts.write_settings()
        self.save_build_id()

    def check_project_modified(self):
        if self.task:
            with self.admin.lock('$code_updating'):
                params = consts.read_params(['CLIENT_MODIFIED', 'SERVER_MODIFIED', 'MAINTENANCE'])
                if not params['MAINTENANCE'] and (params['CLIENT_MODIFIED'] or params['SERVER_MODIFIED']):
                    self.__task_locked = False
                    try:
                        if params['SERVER_MODIFIED']:
                            reload_task(self)
                            consts.BUILD_VERSION += 1
                        if params['CLIENT_MODIFIED']:
                            update_events_code(self.admin)
                        consts.CLIENT_MODIFIED = False
                        consts.SERVER_MODIFIED = False
                        consts.MODIFICATION += 1
                        consts.PARAMS_VERSION += 1
                        consts.write_settings()
                        self.save_build_id()
                    finally:
                        self.__task_locked = True

    @property
    def build_id(self):
        return '%s_%s_%s' % (self.build_id_prefix, consts.BUILD_VERSION, consts.PARAMS_VERSION)

    def save_build_id(self):
        with self.admin.lock('$save_build_id'):
            path = os.path.join(self.work_dir, 'locks')
            for file_name in os.listdir(path):
                if file_name.startswith(self.build_id_prefix):
                    os.remove(os.path.join(path, file_name))
            file_write(os.path.join(path, self.build_id), '')

    def check_build(self):
        path = os.path.join(self.work_dir, 'locks')
        if not os.path.exists(os.path.join(path, self.build_id)):
            with self.admin.lock('$build_checking'):
                cur_build_version = consts.BUILD_VERSION
                cur_params_version = consts.PARAMS_VERSION
                build_version = None
                params_version = None
                for file_name in os.listdir(path):
                    if file_name.startswith(self.build_id_prefix):
                        cur_build_id = file_name
                        arr = cur_build_id.split('_')
                        build_version = int(arr[1])
                        params_version = int(arr[2])
                        break
                if params_version != cur_params_version:
                    consts.read_settings()
                if build_version != cur_build_version:
                    self.__task_locked = False
                    try:
                        reload_task(self)
                    finally:
                        self.__task_locked = True
                    consts.read_settings()

    def on_api(self, request):
        error = ''
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            try:
                data = request.get_data()
                if type(data) != str:
                    data = to_str(data, 'utf-8')
                method, task_id, item_id, params, modification = json.loads(data)
                if task_id == 0:
                    if self.production:
                        raise Exception('Server is in production mode.')
                    task = self.admin
                else:
                    task = self.task
                request.task = task
                self.check_build()
                result = {'status': consts.RESPONSE, 'data': None, 'modification': consts.MODIFICATION}
                if not self.check_session(request, task):
                    result['status'] = consts.PROJECT_NOT_LOGGED
                elif task_id and modification and modification != consts.MODIFICATION:
                    result['status'] = consts.PROJECT_MODIFIED
                elif self.under_maintenance:
                    result['status'] = consts.PROJECT_MAINTAINANCE
                elif method == 'logout':
                    self.logout(request, task)
                    result['status'] = consts.PROJECT_NOT_LOGGED
                else:
                    item = task
                    if task and item_id:
                        item = task.item_by_ID(item_id)
                    self._busy += 1
                    try:
                        data = self.get_response(item, method, params)
                    finally:
                        self._busy -= 1
                    result['data'] = data
                r ['result'] = result
            except AbortException as e:
                traceback.print_exc()
                error = error_message(e)
                r['result'] = {'data': [None, error]}
                r['error'] = error
            except Exception as e:
                self.log.exception(error_message(e))
                error = error_message(e)
                if consts.DEBUGGING and task_id != 0:
                    raise
                r['result'] = {'data': [None, error]}
                r['error'] = error
            response = self.create_post_response(request, r)
            request.save_client_cookie(response, task)
            return response

    def get_response(self, item, method, params):
        if method == 'open':
            return item.select_records(params, client_request=True)
        elif method == 'apply':
            return item.apply_changes(params, safe=True)
        elif method == 'server':
            return self.server_func(item, params[0], params[1])
        elif method == 'print':
            return item.print_report(*params, safe=True), ''
        elif method == 'load':
            return self.init_client(item)

    def server_func(self, obj, func_name, params):
        result = None
        error = ''
        func = getattr(obj, func_name)
        if func:
            result = func(obj, *params)
        else:
            raise Exception('item: %s no server function with name %s' % (obj.item_name, func_name))
        return result, error

    def on_ext(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            method = get_path_info(request.environ)
            data = request.get_data()
            if type(data) != str:
                data = to_str(data, 'utf-8')
            try:
                params = json.loads(data)
            except:
                params = None
            if self.task:
                try:
                    data = None
                    if self.under_maintenance:
                        status = consts.UNDER_MAINTAINANCE
                    elif self.task.on_ext_request:
                        status = consts.RESPONSE
                        self._busy += 1
                        try:
                            data = self.task.on_ext_request(self.task, method, params)
                        finally:
                            self._busy -= 1
                    else:
                        status = None
                    r['result'] = {'status': status, 'data': data, 'modification': consts.MODIFICATION}
                except AbortException as e:
                    traceback.print_exc()
                    r['result'] = {'data': [None, error_message(e)]}
                    r['error'] = error_message(e)
                except Exception as e:
                    traceback.print_exc()
                    r['result'] = {'data': [None, error_message(e)]}
                    r['error'] = error_message(e)
            else:
                r['result'] = {'status': self.state, 'data': None, 'modification': None}
            return self.create_post_response(request, r)

    def on_upload(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            task_id = int(request.form.get('task_id'))
            item_id = int(request.form.get('item_id'))
            field_id = int(request.form.get('field_id'))
            path = request.form.get('path')
            if task_id == 0:
                task = self.admin
            else:
                task = self.task
            if task:
                request.task = task
                result = {'status': consts.RESPONSE, 'data': None, 'modification': consts.MODIFICATION}
                r ['result'] = result
                if not self.check_session(request, task):
                    r['result']['status'] = consts.NOT_LOGGED
                    r['result']['data'] = consts.NOT_LOGGED
                else:
                    f = request.files.get('file')
                    file_name = request.form.get('file_name')
                    if f and file_name:
                        base, ext = os.path.splitext(file_name)
                        ext = ext.lower()
                        upload_result = None
                        if task.on_upload:
                             upload_result = task.on_upload(task, path, file_name, f)
                        if upload_result:
                            path, file_name = upload_result
                            r['result']['data'] = {'file_name': file_name, 'path': path}
                        else:
                            if item_id != -1 and field_id != -1:
                                item = task.item_by_ID(item_id)
                                field = item.field_by_ID(field_id)
                                if field.data_type == consts.IMAGE:
                                    if not valid_uploaded_file('image/*', ext):
                                        r['error'] = consts.lang['upload_not_allowed']
                                elif field.data_type == consts.FILE:
                                    if not valid_uploaded_file(field.field_file['accept'], ext):
                                        r['error'] = consts.lang['upload_not_allowed']
                                else:
                                    r['error'] = 'Operation prohibited'
                            else:
                                if not ext in consts.upload_file_ext:
                                    r['error'] = '%s - %s' % (request.form.get('file_name'), consts.lang['upload_not_allowed'])
                                    self.log.error(r['error'])
                            file_name = secure_filename(file_name)
                            file_name = file_name.replace('?', '')
                            base, ext = os.path.splitext(file_name)
                            date_suffix = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S_%f')
                            file_name = ('%s_%s%s') % (base, date_suffix, ext)
                            if len(file_name) > 255:
                                base = base[:len(base) - (len(file_name) - 255)]
                                file_name = ('%s_%s%s') % (base, date_suffix, ext)
                            if task_id == 0:
                                path = os.path.join('static', 'builder')
                            else:
                                path = os.path.join('static', 'files')
                            if not r['error']:
                                dir_path = os.path.join(to_str(self.work_dir, 'utf-8'), path)
                                if not os.path.exists(dir_path):
                                    os.makedirs(dir_path)
                                f.save(os.path.join(dir_path, file_name))
                                r['result']['data'] = {'file_name': file_name, 'path': path}
                    else:
                        r['error'] = 'File upload invalid parameters'
            else:
                r['result'] = {'status': self.state, 'data': None, 'modification': None}
            return self.create_post_response(request, r)

    def stop(self, sigvalue):
        self.kill()

    def kill(self):
        import signal, subprocess
        if os.name == "nt":
            subprocess.Popen("taskkill /F /T /pid %i" % self.pid, shell=True)
        else :
            os.killpg(self.pid, signal.SIGKILL)



import os
import sys
import datetime
import zipfile
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
from subprocess import Popen, STDOUT, PIPE
import uuid
if os.name == "nt":
    try:
        from winreg import OpenKey, QueryValue, HKEY_LOCAL_MACHINE
    except:
        from _winreg import OpenKey, QueryValue, HKEY_LOCAL_MACHINE

from .common import to_bytes

def tobytes(x):
    return to_bytes(x, 'utf-8')

class Var(object):
    def __init__(self, literal):
        self.literal = tobytes(literal)
        self.var = tobytes(literal[2:-2])

class Cell(object):
    def __init__(self, col, default_formating):
        self.vars = []
        self.col = col
        self.default_formating = default_formating
        self.image = None

class Col(object):
    def __init__(self, repeated):
        if not repeated:
            repeated = 1
        self.repeated = int(repeated)
        self.text = ''

class Row(object):
    def __init__(self, repeated):
        if not repeated:
            repeated = 1
        self.repeated = int(repeated)
        self.cells = []

class Band(object):
    def __init__(self, report, tag=None):
        self.report = report
        self.tag = tag
        self.rows = []
        self.cols = []
        self.text = ''

class Report(object):

    TABLE_END = tobytes('</table:table>')
    START_ROW = tobytes('<table:table-row')
    END_ROW = tobytes('</table:table-row>')
    START_COLUMN = tobytes('<table:table-column')
    END_COLUMN = tobytes('/>')
    START_TEXT = tobytes('<text:p>')
    END_TEXT = tobytes('</text:p>')
    START_PARAM = tobytes('%(')
    END_PARAM = tobytes(')s')

    def prepare_report(self, on_generate, ext='.ods', hidden_columns=None):
        self.hidden_columns = hidden_columns
        self.ext = ext
        file_name = self.gen_file_name()
        self.report_filename = os.path.join(self.dest_folder, file_name)
        self.report_url = os.path.join(self.dest_url, file_name)

        self.header = Band(self)
        self.columns = Band(self)
        self.bands = []
        self.footer = Band(self)
        self.parse()

        self.images = []
        self.cur_row = 1
        self.content_xml = tobytes('')
        self.generate_report(on_generate)
        self.content_xml = None

    def generate_report(self, on_generate):
        if self.on_before_generate:
            self.on_before_generate(self)
        if self.on_parsed:
            self.on_parsed(self)
        self.content_xml += self.header.text
        for col in self.columns.cols:
            self.content_xml += col.text
        if on_generate:
            on_generate(self)
        self.content_xml += self.footer.text
        self.save()
        if self.ext != '.ods':
            self.convert_report()
        if self.on_after_generate:
            self.on_after_generate(self)

    def save(self):
        image_links = self.get_image_links()
        with zipfile.ZipFile(self.report_filename, 'w', zipfile.ZIP_DEFLATED) as dest:
            with zipfile.ZipFile(self.template_path, 'r') as source:
                manifest_data = None
                for file_name in source.namelist():
                    data = source.read(file_name)
                    if file_name == 'content.xml':
                        dest.writestr(file_name, self.content_xml)
                    elif file_name == 'META-INF/manifest.xml':
                        manifest_data = data
                    elif image_links.get(file_name):
                        pass
                    else:
                        dest.writestr(file_name, data)
                if self.images:
                    for new_link, image, link in self.images:
                        dest.write(image, new_link, compress_type = zipfile.ZIP_DEFLATED)
                    manifest_data = self.change_manifest(manifest_data, image_links)
                if manifest_data:
                    dest.writestr('META-INF/manifest.xml', manifest_data)

    def get_image_links(self):
        result = {}
        for new_link, image, link in self.images:
            l = result.get(link)
            if l is None:
                result[link] = []
                l = result[link]
            l.append(new_link)
        return result

    def change_manifest(self, manifest_data, image_links):
        for old_link, new_links in image_links.items():
            old_link = tobytes(old_link)
            pos = manifest_data.find(old_link)
            if pos != -1:
                pos = manifest_data.rfind(tobytes('<manifest:file-entry'), 0, pos)
                pos1 = manifest_data.find(tobytes('/>'), pos)
                old_line = manifest_data[pos:pos1+2]
                new_lines = tobytes('')
                for link in new_links:
                    link = tobytes(link)
                    new_lines += old_line.replace(old_link, link, 1)
                manifest_data = manifest_data.replace(old_line, new_lines, 1)
        return manifest_data

    def convert_report(self):
        converted = False
        if self.on_convert:
            converted = self.on_convert(self)
        else:
            converted = self.convert()
        if converted:
            converted_file_name = self.report_filename.replace('.ods', self.ext)
            if os.path.exists(converted_file_name):
                os.remove(self.report_filename)
                self.report_filename = converted_file_name
                self.report_url = self.report_url.replace('.ods', self.ext)

    def convert(self):
        with self.task.lock('$report_conversion'):
            try:
                from subprocess import Popen, STDOUT, PIPE
                if os.name == "nt":
                    regpath = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\soffice.exe"
                    root = OpenKey(HKEY_LOCAL_MACHINE, regpath)
                    s_office = QueryValue(root, "")
                else:
                    s_office = "soffice"
                convertion = Popen([s_office, '--headless', '--convert-to', self.ext.strip('.'),
                    self.report_filename, '--outdir', self.dest_folder],
                    stderr=STDOUT,stdout=PIPE)
                out, err = convertion.communicate()
            except Exception as e:
                print('Report "%s" conversion error:' % self.name, e)
                return False
            return True

    def gen_file_name(self):
        file_name = escape(self.name, { char: '' for char in ['\\', '/', ':', '*', '"', '<', '>', '|', 'NUL']})
        file_name = file_name + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S.%f') + '.ods'
        return file_name

    def parse(self):
        with zipfile.ZipFile(self.template_path, 'r') as z:
            content = z.read('content.xml')
        dom = parseString(content)
        tables = dom.getElementsByTagName('table:table')
        table = tables[0]
        cur_band = self.header
        self.cur_rows = []
        for child in table.childNodes:
            if child.nodeName == 'table:table-column':
                repeated = child.getAttribute('table:number-columns-repeated')
                self.columns.cols.append(Col(repeated))
            elif child.nodeName == 'table:table-row':
                for row_child in child.childNodes:
                    if row_child.nodeName == 'table:table-cell':
                        text = row_child.getElementsByTagName('text:p')
                        if text.length > 0:
                            tag = text[0].childNodes[0].nodeValue
                            cur_band.rows = self.cur_rows[:]
                            self.cur_rows = []
                            cur_band = Band(self, tag)
                            self.bands.append(cur_band)
                        break
            self.process_rows(child)
        cur_band.rows = self.cur_rows[:]
        if not len(self.bands):
            raise Exception('No bands in the report template "%s"' % self.template_path)
        self.process_bands(content)
        self.bands_dict = { band.tag: band for band in self.bands }

    def process_bands(self, content):
        start = content.find(self.START_COLUMN)
        self.header.text = content[:start]
        for col in self.columns.cols:
            start = content.find(self.START_COLUMN, start)
            end = content.find(self.END_COLUMN, start) + len(self.END_COLUMN)
            col.text = content[start:end]
            start = end
        band_positions = []
        for band in self.bands:
            tag = self.START_TEXT + tobytes(band.tag) + self.END_TEXT
            tag_start = content.find(tag, start)
            start = content.rfind(self.START_ROW, start, tag_start)
            band_positions.append(start)
        end = None
        for row in band.rows:
            if row.repeated > 1000:
                tag = tobytes('table:number-rows-repeated="%s"' % row.repeated)
                tag_start = content.find(tag, start)
                end = content.rfind(self.START_ROW, start, tag_start)
        if not end:
            tag_start = content.find(self.TABLE_END, start)
            end = content.rfind(self.END_ROW, start, tag_start)
            end += len(self.END_ROW)
        band_positions.append(end)
        for i, band in enumerate(self.bands):
            band.text = content[band_positions[i]:band_positions[i+1]]
            band.text = band.text.replace(self.START_TEXT + tobytes(band.tag) + self.END_TEXT, tobytes(''), 1)
        self.footer.text = content[end:]

    def find_vars(self, text):
        result = []
        while True:
            p1 = text.find('%(')
            if p1 != -1:
                p2 = text.find(')s', p1)
                if p2 != -1:
                    result.append(Var(text[p1:p2+2]))
                text = text[p1+2:]
            else:
                break
        return result

    def process_rows(self, node):
        if node.nodeName == 'table:table-row':
            repeated = node.getAttribute('table:number-rows-repeated')
            row = Row(repeated)
            col_count = 1
            for child in node.childNodes:
                if child.nodeName == 'table:table-cell':
                    text = child.getElementsByTagName('text:p')
                    cell = None
                    for t in text:
                        for child_node in t.childNodes:
                            cell_text = child_node.nodeValue
                            if cell_text:
                                trimed = cell_text.strip()
                                cell_vars = self.find_vars(trimed)
                                if cell_vars:
                                    default = t == text[0] and child_node == t.childNodes[0]
                                    if not cell:
                                        cell = Cell(col_count, default)
                                        row.cells.append(cell)
                                    for v in cell_vars:
                                        cell.vars.append(v)
                    draw_frame = child.getElementsByTagName('draw:frame')
                    if draw_frame.length and cell:
                        draw_image = child.getElementsByTagName('draw:image')
                        if draw_image.length:
                            cell.image = {
                                'table:end-cell-address': draw_frame[0].attributes['table:end-cell-address'].value,
                                'xlink:href': draw_image[0].attributes['xlink:href'].value
                            }
                    col_count += 1
                elif child.nodeName == 'table:covered-table-cell':
                    cols = child.getAttribute('table:number-columns-repeated')
                    if cols:
                        col_count += int(cols)
            self.cur_rows.append(row)
        else:
            for child in node.childNodes:
                self.process_rows(child)

    def print_band(self, band_tag, var_dict=None):
        band = self.bands_dict[band_tag]
        cell_vars = {}
        if var_dict:
            for key, value in var_dict.items():
                cell_vars[tobytes(key)] = value
        band_text = band.text
        for row in band.rows:
            for cell in row.cells:
                image_processed = False
                for var in cell.vars:
                    value = cell_vars.get(var.var)
                    image = None
                    if not value is None:
                        if isinstance(value, str):
                            text = tobytes(escape(value).replace('\n', '</text:p><text:p>'))
                        elif isinstance(value, int) or type(value) == float:
                            text = tobytes(str(value))
                        elif type(value) == dict and cell.image:
                            image = value.get('image')
                            text = value.get('text', '')
                        else:
                            raise Exception('Invalid cell value type')
                        band_text = band_text.replace(var.literal, tobytes(text), 1)
                        if image:
                            image_processed = True
                            if os.path.exists(image):
                                addres = cell.image['table:end-cell-address']
                                new_addres = '%s.%s%s' % (addres.split('.')[0], self.col_int_to_str(cell.col), self.cur_row)
                                band_text = band_text.replace(tobytes(addres), tobytes(new_addres), 1)
                                link = cell.image['xlink:href']
                                new_link = 'Pictures/%s%s' % (self.cur_row, self.gen_image_name(image))
                                band_text = band_text.replace(tobytes(link), tobytes(new_link), 1)
                                self.images.append([new_link, image, link])
                            else:
                                link = cell.image['xlink:href']
                                band_text = band_text.replace(tobytes(link), tobytes(''), 1)
                if cell.image and not image_processed:
                    link = cell.image['xlink:href']
                    band_text = band_text.replace(tobytes(link), tobytes(''), 1)
            self.cur_row += row.repeated
        self.content_xml += band_text

    def gen_image_name(self, image_path):
        return uuid.uuid4().hex + os.path.splitext(image_path)[1]

    def hide_columns(self, hidden_columns):
        if hidden_columns:
            cols = []
            for col in self.columns.cols:
                if col.repeated > 1:
                    parts = col.text.split(tobytes(' '))
                    text = tobytes(' ').join([ part for part in parts if part.find(tobytes('table:number-columns-repeated')) == -1])
                    for i in range(col.repeated):
                        new_col = Col(1)
                        new_col.text = text
                        cols.append(new_col)
                else:
                    cols.append(col)
            hidden_ints = [ self.col_str_to_int(c) for c in hidden_columns ]
            for i, col in enumerate(cols):
                if i + 1 in hidden_ints:
                    col.text = col.text[0:-2] + tobytes(' table:visibility="collapse"/>')
            self.columns.cols = []
            for col in cols:
                self.columns.cols.append(col)

    def col_str_to_int(self, s):
        if type(s) == int:
            return s
        s = s.upper()
        base = ord('A')
        mult = ord('Z') - base + 1
        result = 0
        chars = [ s[i] for i in range(len(s))]
        for i, char in enumerate(reversed(chars)):
            result += (ord(chars[i]) - base + 1) * (mult ** (len(chars) - i - 1))
        return result

    def col_int_to_str(self, i):
        base = ord('A')
        mult = ord('Z') - base + 1
        result = ''
        d = i
        while True:
            d -= 1
            d, m = divmod(d, mult)
            c = chr(m + base)
            result = c + result
            if d == 0:
                break
        return result



import pymssql

from .mssql_db import MSSqlDB

class MSSqlDB1(MSSqlDB):
    def __init__(self):
        MSSqlDB.__init__(self)

    def connect(self, db_info):
        if db_info.encoding:
            return pymssql.connect(server=db_info.server, database=db_info.database,
                user=db_info.user, password=db_info.password, host=db_info.host,
                port=db_info.port, charset=db_info.encoding, autocommit=True)
        else:
            return pymssql.connect(server=db_info.server, database=db_info.database,
                user=db_info.user, password=db_info.password, host=db_info.host,
                port=db_info.port, autocommit=True)

    def get_lastrowid(self, cursor):
        return cursor.lastrowid

    def value_literal(self, index):
        return '%s'

db = MSSqlDB1()




from ..common import consts

def get_database(app, db_type, lib):
    db = None
    if db_type == consts.SQLITE:
        from .sqlite_db import db
    elif db_type == consts.POSTGRESQL:
        from .postgres_db import db
    elif db_type == consts.MYSQL:
        if lib == 1:
            from .mysql_db1 import db
        elif lib == 2:
            from .mysql_db2 import db
        else:
            from .mysql_db import db
    elif db_type == consts.FIREBIRD:
        from .firebird_db import db
    elif db_type == consts.ORACLE:
        from .oracle_db import db
    elif db_type == consts.MSSQL:
        if lib == 1:
            from .mssql_db1 import db
        elif lib == 2:
            from .mssql_db2 import db
        else:
            from .mssql_db import db
    if db:
        db.app = app
    return db



import pyodbc

from .mssql_db import MSSqlDB

class MSSqlDB2(MSSqlDB):
    def __init__(self):
        MSSqlDB.__init__(self)

    #DRIVER={ODBC Driver 17 for SQL Server};SERVER=ANDREW-PC\SQLEXPRESS;DATABASE=demo;UID=sa;PWD=1111
    #DSN=SQL Server;DATABASE=demo;UID=sa;PWD=1111
    def connect(self, db_info):
        return pyodbc.connect(db_info.dns)

    def get_lastrowid(self, cursor):
        cursor.execute('SELECT @@IDENTITY')
        r = cursor.fetchone()
        return int(r[0])

    @property
    def arg_params(self):
        return True

db = MSSqlDB2()




import psycopg2
psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)

from ..common import consts
from .db import AbstractDB

class PostgresDB(AbstractDB):
    def __init__(self):
        AbstractDB.__init__(self)
        self.db_type = consts.POSTGRESQL
        self.DDL_ROLLBACK = True
        self.DESC_NULLS = 'NULLS LAST'
        self.ASC_NULLS = 'NULLS FIRST'
        self.LIKE = 'ILIKE'
        self.FIELD_TYPES = {
            consts.INTEGER: 'INTEGER',
            consts.TEXT: 'TEXT',
            consts.FLOAT: 'NUMERIC',
            consts.CURRENCY: 'NUMERIC',
            consts.DATE: 'DATE',
            consts.DATETIME: 'TIMESTAMP',
            consts.BOOLEAN: 'INTEGER',
            consts.LONGTEXT: 'TEXT',
            consts.KEYS: 'TEXT',
            consts.FILE: 'TEXT',
            consts.IMAGE: 'TEXT'
        }

    def get_params(self, lib):
        params = self.params
        params['name'] = 'POSTGRESQL'
        params['login'] = True
        params['password'] = True
        params['encoding'] = True
        params['host'] = True
        params['port'] = True
        params['dsn'] = True
        return params

    def connect(self, db_info):
        if db_info.dsn:
            return psycopg2.connect(dsn=db_info.dsn)
        else:
            return psycopg2.connect(dbname=db_info.database, user=db_info.user,
                password=db_info.password, host=db_info.host, port=db_info.port,
                client_encoding=db_info.encoding)

    def get_select(self, query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
        start = fields_clause
        end = ''.join([from_clause, where_clause, group_clause, order_clause])
        offset = query.offset
        limit = query.limit
        result = 'SELECT %s FROM %s' % (start, end)
        if limit:
            result += ' LIMIT %d OFFSET %d' % (limit, offset)
        return result

    def value_literal(self, index):
        return '%s'

    def convert_like(self, field_name, val, data_type):
        return '%s::text' % field_name, val.upper()

    def create_table(self, table_name, fields, gen_name=None, foreign_fields=None):
        sql = ''
        primary_key = ''
        seq_name = gen_name
        sql = 'CREATE TABLE "%s"\n(\n' % table_name
        lines = []
        for field in fields:
            default_text = self.default_text(field)
            if field.primary_key:
                field_type = 'SERIAL PRIMARY KEY'
            else:
                field_type = self.FIELD_TYPES[field.data_type]
            line = '"%s" %s' % (field.field_name, field_type)
            if not default_text is None:
                line += ' DEFAULT %s' % default_text
            lines.append(line)
        sql += ',\n'.join(lines)
        sql += ')\n'
        return sql

    def drop_table(self, table_name, gen_name):
        result = []
        result.append('DROP TABLE "%s"' % table_name)
        if gen_name:
            result.append('DROP SEQUENCE IF EXISTS "%s"' % gen_name)
        return result

    def add_field(self, table_name, field):
        default_text = self.default_text(field)
        line = 'ALTER TABLE "%s" ADD COLUMN "%s" %s' % \
            (table_name, field.field_name, self.FIELD_TYPES[field.data_type])
        if not default_text is None:
            line += ' DEFAULT %s' % default_text
        return line

    def del_field(self, table_name, field):
        return 'ALTER TABLE "%s" DROP COLUMN "%s"' % (table_name, field.field_name)

    def change_field(self, table_name, old_field, new_field):
        result = []
        default_text = self.default_text(new_field)
        field_info = self.get_field_info(old_field.field_name, table_name)
        if old_field.field_name != new_field.field_name:
            result.append('ALTER TABLE "%s" RENAME COLUMN  "%s" TO "%s"' % \
                (table_name, old_field.field_name, new_field.field_name))
        if old_field.size != new_field.size:
            if field_info['data_type'].upper() in ['CHARACTER VARYING', 'VARCHAR', 'CHARACTER', 'CHAR'] and \
                field_info['size'] < new_field.size:
                line = 'ALTER TABLE "%s" ALTER COLUMN "%s" TYPE %s(%d) ' % \
                    (table_name, new_field.field_name, field_info['data_type'], new_field.size)
                result.append(line)
        if old_field.default_value != new_field.default_value:
            if not default_text is None:
                line = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT %s' % \
                    (table_name, new_field.field_name, default_text)
            else:
                line = 'ALTER TABLE "%s" ALTER "%s" DROP DEFAULT' % \
                    (table_name, new_field.field_name)
            result.append(line)
        return result

    def create_index(self, index_name, table_name, unique, fields, desc):
        return 'CREATE %s INDEX "%s" ON "%s" (%s)' % (unique, index_name, table_name, fields)

    def drop_index(self, table_name, index_name):
        return 'DROP INDEX "%s"' % index_name

    def create_foreign_index(self, table_name, index_name, key, ref, primary_key):
        return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s") MATCH SIMPLE' % \
            (table_name, index_name, key, ref, primary_key)

    def drop_foreign_index(self, table_name, index_name):
        return 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % (table_name, index_name)

    def insert_query(self, pk_field):
        return 'INSERT INTO "%s" (%s) VALUES (%s) RETURNING ' + pk_field.db_field_name

    def before_insert(self, cursor, pk_field):
        if pk_field and pk_field.owner.gen_name and not pk_field.data:
            cursor.execute(self.next_sequence(pk_field.owner.gen_name))
            rows = cursor.fetchall()
            pk_field.data = rows[0][0]

    def after_insert(self, cursor, pk_field):
        if pk_field and not pk_field.owner.gen_name and not pk_field.data:
            pk_field.data = cursor.fetchone()[0]

    def next_sequence(self, gen_name):
        return 'SELECT NEXTVAL(\'"%s"\')' % gen_name

    def restart_sequence(self, gen_name, value):
        return 'ALTER SEQUENCE "%s" RESTART WITH %d' % (gen_name, value)

    def identifier_case(self, name):
        return name.lower()

    def get_table_names(self, connection):
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema'")
        result = cursor.fetchall()
        return [r[1] for r in result]

    def get_table_info(self, connection, table_name, db_name):
        cursor = connection.cursor()
        sql = "select column_name, data_type, character_maximum_length, column_default from information_schema.columns where table_name='%s'" % table_name
        cursor.execute(sql)
        result = cursor.fetchall()
        fields = []
        for column_name, data_type, character_maximum_length, column_default in result:
            try:
                size = 0
                if character_maximum_length:
                    size = character_maximum_length
                default_value = None
                if column_default:
                    default_value = column_default.split('::')[0]
                    if default_value.find('nextval') != -1:
                        default_value = None
            except:
                pass
            pk = False
            fields.append({
                'field_name': column_name,
                'data_type': data_type.upper(),
                'size': size,
                'default_value': column_default,
                'pk': pk
            })
        return {'fields': fields, 'field_types': self.FIELD_TYPES}

db = PostgresDB()



import sys, os
import zipfile
import datetime, time
import inspect
import pickle
import json
import types

from werkzeug.security import generate_password_hash, check_password_hash
from werkzeug.utils import redirect
from sqlalchemy.pool import NullPool, QueuePool

from .filelock import FileLock

from .common import consts, error_message, json_defaul_handler, QueryData
from .common  import to_bytes
from .db.databases import get_database
from .tree import AbstrTask, AbstrGroup, AbstrItem, AbstrDetail, AbstrReport
from .dataset import Dataset, DBField, DBFilter, ParamReport, Param, DatasetException
from .admin.copy_db import copy_database
from .report import Report

class ServerDataset(Dataset):
    def __init__(self):
        Dataset.__init__(self)
        self.ID = None
        self.gen_name = None
        self._order_by = []
        self.values = None
        self.on_open = None
        self.on_before_open = None
        self.on_after_open = None
        self.on_apply = None
        self.on_before_apply_record = None
        self.on_after_apply_record = None
        self.on_count = None
        self.on_field_get_text = None

    def copy(self, filters=True, details=True, handlers=True):
        if self.master:
            raise DatasetException(u'A detail item can not be copied: %s' % self.item_name)
        result = self._copy(filters, details, handlers)
        return result

    def free(self):
        try:
            for d in self.details:
                d.__dict__ = {}
            for f in self.filters:
                f.field = None
                f.__dict__ = {}
            self.filters.__dict__ = {}
            self.__dict__ = {}
        except:
            pass

    def _copy(self, filters=True, details=True, handlers=True):
        result = super(ServerDataset, self)._copy(filters, details, handlers)
        result.table_name = self.table_name
        result.gen_name = self.gen_name
        result._order_by = self._order_by
        result.soft_delete = self.soft_delete
        result._primary_key = self._primary_key
        result._deleted_flag = self._deleted_flag
        result._record_version = self._record_version
        result._master_id = self._master_id
        result._master_rec_id = self._master_rec_id
        result._master_field = self._master_field
        result._master_field_db_field_name = self._master_field_db_field_name
        result._primary_key_db_field_name = self._primary_key_db_field_name
        result._deleted_flag_db_field_name = self._deleted_flag_db_field_name
        result._master_id_db_field_name = self._master_id_db_field_name
        result._master_rec_id_db_field_name = self._master_rec_id_db_field_name
        result._record_version_db_field_name = self._record_version_db_field_name
        return result

    def get_event(self, caption):
        return getattr(caption)

    def add_field(self, *args, **kwargs):
        field_def = self.add_field_def(*args, **kwargs)
        field = DBField(self, field_def)
        self._fields.append(field)
        return field

    def create_fields(self, info):
        for field_def in info:
            self.field_defs.append(field_def)
            field = DBField(self, field_def)
            self._fields.append(field)

    def add_filter(self, *args):
        filter_def = self.add_filter_def(*args)
        fltr = DBFilter(self, filter_def)
        self.filters.append(fltr)
        return fltr

    def create_filters(self, info):
        for filter_def in info:
            self.add_filter(*filter_def)

    def do_internal_open(self, params, connection):
        return self.select_records(params, connection)

    def do_apply(self, params=None, safe=False, connection=None):
        if not self.master:
            changes = {}
            if not params:
                params = {}
            if self.change_log and self.change_log.get_changes(changes):
                data, error = self.apply_changes((changes, params), safe, connection)
                if error:
                    raise Exception(error)
                else:
                    self.change_log.update(data)

    def add_detail(self, master, master_field):
        detail = Detail(self.task, self, master.item_name, master.item_caption,
            master.table_name, master_field=master_field)
        detail.prototype = master
        self.details.append(detail)
        detail.owner = self
        detail.init_fields()
        return detail

    def detail_by_name(self, caption):
        for table in self.details:
            if table.item_name == caption:
                return table

    def _execute_select(self, sql, params):
        db = self.task.db
        con = self.task.connect()
        cursor = con.cursor()
        return self.__execute_select(cursor, sql, params, db)

    def __execute_select(self, cursor, sql, params, db):
        try:
            if params:
                params = db.process_query_params(params, cursor)
                cursor.execute(sql, params)
            else:
                cursor.execute(sql)
            return db.process_query_result(cursor.fetchall())
        except Exception as x:
            self.log.exception('%s:\n%s' % (error_message(x), sql))
            print(sql)
            print(params)

    def execute_open(self, params, connection=None, db=None, query_data=None, ):
        rows = None
        error_mes = ''
        if not db:
            db = self.task.db
        if not query_data:
            query_data = QueryData(params)
        limit = query_data.limit
        offset = query_data.offset
        sqls = db.get_select_queries(self, query_data)
        if connection:
            con = connection
        else:
            con = self.task.connect()
        cursor = con.cursor()
        try:
            if len(sqls) == 1:
                rows = self.__execute_select(cursor, sqls[0][0], sqls[0][1], db)
            else:
                rows = []
                cut = False
                for sql in sqls:
                    rows += self.__execute_select(cursor, sql[0], sql[1], db)
                    if limit or offset:
                        if len(rows) >= offset + limit:
                            rows = rows[offset:offset + limit]
                            cut = True
                            break
                if (limit or offset) and not cut:
                    rows = rows[offset:offset + limit]
                if query_data.summary:
                    new_rows = [0 for r in rows[0]]
                    for row in rows:
                        for i, r in enumerate(row):
                            if not r is None:
                                new_rows[i] += r
                    rows = [new_rows]
        finally:
            if not connection:
                con.close()
        return rows, error_mes

    def init_open_dataset(self, query_data, dataset, result):
        if not dataset:
            if self.master:
                dataset = self.prototype.copy(filters=False, details=False, handlers=False)
            else:
                dataset = self.copy(filters=False, details=False, handlers=False)
            dataset.log_changes = False
            dataset.open(expanded=query_data.expanded, fields=query_data.fields, open_empty=True)
            dataset._dataset = result
            dataset.first();
        return dataset

    def select_records(self, params, connection=None, client_request=False):
        if client_request and not self.can_view():
            raise Exception(consts.language('cant_view') % self.item_caption)

        query_data = QueryData(params)
        query_data.client_request = client_request

        result = None # on_open is depricated
        if self.task.on_open:
            result = self.task.on_open(self, params)
        if result is None and self.on_open:
            result = self.on_open(self, params)
        if result:
            return result

        error = None
        result = None
        exec_query = True
        if self.task.on_before_open:
            exec_query = self.task.on_before_open(self, query_data, params, connection)
        if exec_query != False and self.on_before_open:
            exec_query = self.on_before_open(self, query_data, params, connection)

        if exec_query != False:
            if self.virtual_table:
                result = []
            else:
                result, error = self.execute_open(params, connection, query_data=query_data)

        dataset = None
        if self.on_after_open:
            dataset = self.init_open_dataset(query_data, dataset, result)
            self.on_after_open(self, query_data, params, connection, dataset)
        elif self.task.on_after_open:
            dataset = self.init_open_dataset(query_data, dataset, result)
            self.task.on_after_open(self, query_data, params, connection, dataset)
        if dataset:
            result = dataset._dataset
        return result, error

    def apply_delta(self, delta, params, connection, db=None):
        if not db:
            db = self.task.db
        for d in delta:
            if not d.rec_deleted():
                d.check_record_valid()
        db.process_changes(delta, connection, params)
        return delta.change_log.prepare_updates()

    def set_apply_connection(self, delta, con):
        delta._apply_connection = con
        for d in delta.details:
            self.set_apply_connection(d, con)

    def _get_delta_params(self, delta, params_dict):
        return params_dict.get(str(delta.ID), {})

    def apply_changes(self, data, safe, connection=None):
        result = None
        result_data = None
        changes, params_dict = data
        delta = self.delta(changes, safe)
        delta_params = self._get_delta_params(delta, params_dict)
        if connection:
            con = connection
        else:
            con = self.task.connect()
        self.set_apply_connection(delta, con)
        try:
            if self.task.on_apply:
                result = self.task.on_apply(self, delta, delta_params, con)
            if result is None and self.on_apply:
                result = self.on_apply(self, delta, delta_params, con)
            if result is None:
                result = self.apply_delta(delta, params_dict, con)
            if not connection and con:
                con.commit()
        finally:
            self.set_apply_connection(delta, None)
            if not connection and con:
                con.close()
        return result, ''

    def update_deleted(self, details=None, connection=None): #depricated
        if self._is_delta:
            if details is None:
                details = self.details
            rec_no = self.rec_no
            try:
                for it in self:
                    if it.rec_deleted():
                        for detail in details:
                            fields = []
                            for field in detail.fields:
                                fields.append(field.field_name)
                            prototype = self.task.item_by_ID(detail.prototype.ID).copy()
                            where = {
                                prototype._master_id: self.ID,
                                prototype._master_rec_id: self._primary_key_field.value
                            }
                            prototype.open(fields=fields, expanded=detail.expanded, where=where, connection=connection)
                            if prototype.record_count():
                                it.edit()
                                for p in prototype:
                                    detail.append()
                                    for field in detail.fields:
                                        f = p.field_by_name(field.field_name)
                                        field.set_value(f.value, f.lookup_value)
                                    detail.post()
                                it.post()
                                for d in detail:
                                    d.record_status = consts.RECORD_DELETED
            finally:
                self.rec_no = rec_no

    def field_by_id(self, id_value, field_name):
        return self.get_field_by_id((id_value, field_name))

    def get_field_by_id(self, params):
        id_value, fields = params
        if not (isinstance(fields, tuple) or isinstance(fields, list)):
            fields = [fields]
        copy = self.copy()
        copy.set_where(id=id_value)
        copy.open(fields=fields)
        if copy.record_count() == 1:
            result = []
            for field_name in fields:
                result.append(copy.field_by_name(field_name).value)
            if len(fields) == 1:
                return result[0]
            else:
                return result
        return

    def empty(self):
        if not self.master and self.table_name:
            con = self.task.connect()
            try:
                cursor = con.cursor()
                cursor.execute(self.task.db.empty_table_query(self))
                con.commit()
            except Exception as e:
                self.log.exception(error_message(e))
                con.rollback()
            finally:
                con.close()

    def get_next_id(self, db=None):
        if db is None:
            db = self.task.db
        sql = db.next_sequence(self.gen_name)
        if sql:
            rec = self.task.select(sql)
            if rec:
                if rec[0][0]:
                    return int(rec[0][0])

    def load_interface(self):
        self._view_list = []
        self._edit_list = []
        self._order_list = []
        self._reports_list = []
        value = self.f_info.value
        if value:
            if len(value) >= 4 and value[0:4] == 'json':
                lists = json.loads(value[4:])
            else:
                lists = pickle.loads(to_bytes(value, 'utf-8'))
            self._view_list = lists['view']
            self._edit_list = lists['edit']
            self._order_list = lists['order']
            if lists.get('reports'):
                self._reports_list = lists['reports']

    def store_interface(self, connection=None, apply_interface=True):
        handlers = self.store_handlers()
        self.clear_handlers()
        try:
            self.edit()
            dic = {'view': self._view_list,
                    'edit': self._edit_list,
                    'order': self._order_list,
                    'reports': self._reports_list}
            self.f_info.value = 'json' + json.dumps(dic, default=json_defaul_handler)
            self.post()
            if apply_interface:
                self.apply(connection)
        finally:
            handlers = self.load_handlers(handlers)

    def store_index_fields(self, f_list):
        return json.dumps(f_list)

    def load_index_fields(self, value):
        return json.loads(str(value))


class Group(AbstrGroup):
    def __init__(self, task, owner, name='', caption=''):
        AbstrGroup.__init__(self, task, owner, name, caption)
        self.ID = None

    def add_item(self, name, caption):
        result = Item(self.task, self, name, caption)
        return result

    def get_child_class(self):
        return Item

class ReportGroup(Group):
    def __init__(self, task, owner, name='', caption=''):
        Group.__init__(self, task, owner, name, caption)
        self.on_convert_report = None

    def add_report(self, name, caption):
        result = Report(self.task, self, name, caption)
        return result

    def get_child_class(self):
        return Report

class Item(AbstrItem, ServerDataset):
    def __init__(self, task, owner, name='', caption=''):
        item_type_id = consts.ITEM_TYPE
        AbstrItem.__init__(self, task, owner, name, caption)
        ServerDataset.__init__(self)
        self.reports = []

    def get_child_class(self):
        return Detail

    def get_reports_info(self):
        result = []
        for report in self.reports:
            result.append(report.ID)
        return result

    def get_select_statement(self, query, db=None): # depricated
        if db is None:
            db = self.task.db
        return db.get_select_statement(self, query)

    def order_clause(self, query, db=None): # depricated
        if db is None:
            db = self.task.db
        return db.order_clause(self, query)

    def where_clause(self, query, db=None): # depricated
        if db is None:
            db = self.task.db
        return db.where_clause(self, query)


class Report(AbstrReport, ParamReport, Report):
    def __init__(self, task, owner, name='', caption=''):
        AbstrReport.__init__(self, task, owner, name, caption)
        ParamReport.__init__(self)
        self.item_type_id = consts.REPORT_TYPE

        self.on_before_generate = None
        self.on_generate = None
        self.on_after_generate = None
        self.on_parsed = None
        self.on_before_save_report = None
        self.on_field_get_text = None
        self.on_convert_report = None

    def copy(self):
        result = self.__class__(self.task, None, self.item_name, self.item_caption);
        result.template = self.template
        result.on_before_generate = self.on_before_generate
        result.on_generate = self.on_generate
        result.on_after_generate = self.on_after_generate
        result.on_before_save_report = self.on_before_save_report
        result.on_parsed = self.on_parsed
        result.on_convert_report = self.owner.on_convert_report
        result.param_defs = self.param_defs
        for param_def in result.param_defs:
            param = Param(result, param_def)
            result.params.append(param)
        result.prepare_params()
        return  result

    def print_report(self, param_values, url, ext=None, safe=False):
        self.delete_reports()
        if safe and not self.can_view():
            raise Exception(consts.language('cant_view') % self.item_caption)
        copy = self.copy()
        copy.name = self.item_name
        copy.template_path = os.path.join(self.task.work_dir, 'reports', self.template)
        copy.url = url
        copy.dest_folder = os.path.join(self.task.work_dir, 'static', 'reports')
        copy.dest_url = os.path.join(url, 'static', 'reports')
        if not os.path.exists(copy.dest_folder):
            os.makedirs(copy.dest_folder)
        copy.on_convert = self.on_convert_report
        for i, param in enumerate(copy.params):
            param.data = param_values[i];
        copy.prepare_report(self.on_generate, ext='.' + ext)
        if url:
            return copy.report_url

    def delete_reports(self):
        task = self.task
        if consts.DELETE_REPORTS_AFTER:
            path = os.path.join(task.work_dir, 'static', 'reports')
            if os.path.isdir(path):
                for f in os.listdir(path):
                    file_name = os.path.join(path, f)
                    if os.path.isfile(file_name):
                        delta = datetime.datetime.now() - datetime.datetime.fromtimestamp(os.path.getmtime(file_name))
                        hours, sec = divmod(delta.total_seconds(), 3600)
                        if hours > consts.DELETE_REPORTS_AFTER:
                            os.remove(file_name)


class DBInfo(object):
    def __init__(self, dsn='', server = '', lib=None, database = '',
        user = '', password = '', host='', port='', encoding=''):
        self.dsn = dsn
        self.server = server
        self.lib = lib
        self.database = database
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.encoding = encoding
        if not self.port:
            self.port = None
        else:
            self.port = int(port)
        if not self.encoding:
            self.encoding = None


class AbstractServerTask(AbstrTask):
    def __init__(self, app, name, caption):
        AbstrTask.__init__(self, None, None, None)
        self.app = app
        self.items = []
        self.lookup_lists = {}
        self.ID = None
        self.item_name = name
        self.item_caption = caption
        self.visible = True
        self.on_before_request = None
        self.on_after_request = None
        self.on_open = None
        self.on_before_open = None
        self.on_after_open = None
        self.on_apply = None
        self.on_before_apply_record = None
        self.on_after_apply_record = None
        self.on_count = None
        self.on_request = None
        self.on_upload = None
        self.work_dir = app.work_dir
        self.modules = []
        self.log = app.log
        self.consts = consts
        self.pool = None

    @property
    def version(self):
        return consts.VERSION

    def get_child_class(self):
        return Group

    def create_pool(self, con_pool_size=1, persist_con=True):
        if self.pool:
            self.pool.dispose()
        if persist_con:
            if self.db_type == consts.SQLITE:
                self.pool = NullPool(self.create_connection)
            else:
                self.pool = QueuePool(self.create_connection, \
                    pool_size=con_pool_size, max_overflow=con_pool_size*2, \
                    recycle=60*60)
        else:
            self.pool = NullPool(self.create_connection)

    def create_connection(self):
        return self.db.connect(self.db_info)

    def connect(self):
        return self.pool.connect()

    def __execute_query_list(self, cursor, query_list):
        for query in query_list:
            if query:
                if type(query) == list:
                    self.__execute_query_list(cursor, query)
                else:
                    if type(query) == tuple:
                        self.execute_query(cursor, query[0], query[1])
                    else:
                        self.execute_query(cursor, query)

    def execute(self, query, params=None, connection=None, db=None):
        # ~ print (query, params)
        error = None
        con = connection
        if not connection:
            con = self.connect()
        cursor = con.cursor()
        try:
            if type(query) == list:
                self.__execute_query_list(cursor, query)
            else:
                self.execute_query(cursor, query, params)
        except Exception as x:
            error = error_message(x)
        finally:
            if not connection:
                con.commit()
                con.close()
        return error

    def select(self, select_query, connection=None, db=None):
        result = None
        error = None
        con = connection
        if not connection:
            con = self.connect()
        cursor = con.cursor()
        try:
            self.execute_query(cursor, select_query)
            result = cursor.fetchall()
            result = [list(r) for r in result]
        except Exception as x:
            error = error_message(x)
        finally:
            if not connection:
                con.close()
        return result

    def execute_select(self, command): #depricated
        return self.select(command)

    def generate_password_hash(self, password, method='pbkdf2:sha256', salt_length=8):
        return generate_password_hash(password, method, salt_length)

    def check_password_hash(self, pwhash, password):
        return check_password_hash(pwhash, password)

    def get_module_name(self):
        return str(self.item_name)

    def lock(self, lock_name, timeout=-1):
        lock_file = os.path.join(self.work_dir, 'locks', lock_name + '.lock')
        locks_dir = os.path.dirname(lock_file)
        if not os.path.exists(locks_dir):
            os.makedirs(locks_dir)
        return FileLock(lock_file, timeout)

    def compile_item(self, item):
        item.module_name = None
        code = item.server_code
        item.module_name = item.get_module_name()
        item_module = type(sys)(item.module_name)
        item_module.__dict__['task'] = self
        sys.modules[item.module_name] = item_module

        item.task.modules.append(item.module_name)
        if item.owner:
            sys.modules[item.owner.get_module_name()].__dict__[item.module_name] = item_module
        if code:
            try:
                code = to_bytes(code, 'utf-8')
            except Exception as e:
                self.log.exception(error_message(e))
            comp_code = compile(code, item.module_name, "exec")
            exec(comp_code, item_module.__dict__)

            item_module.__dict__['__loader__'] = item._loader
            funcs = inspect.getmembers(item_module, inspect.isfunction)
            item._events = []
            for func_name, func in funcs:
                item._events.append((func_name, func))
                # ~ if hasattr(item, func_name) and func_name[:3] != 'on_':
                    # ~ item.log.warning('Module %s: method "%s" will override "%s" existing attribute. Please, rename the function.' % \
                        # ~ (item.module_name, func_name, item.item_name))
                setattr(item, func_name, func)
        del code

    def convert_report(self, report, ext):
        converted = False
        with self.task.lock('$report_conversion'):
            try:
                from subprocess import Popen, STDOUT, PIPE
                if os.name == "nt":
                    regpath = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\soffice.exe"
                    root = OpenKey(HKEY_LOCAL_MACHINE, regpath)
                    s_office = QueryValue(root, "")
                else:
                    s_office = "soffice"
                convertion = Popen([s_office, '--headless', '--convert-to', ext,
                    report.report_filename, '--outdir', os.path.join(self.work_dir, 'static', 'reports')],
                    stderr=STDOUT,stdout=PIPE)
                out, err = convertion.communicate()
                converted = True
            except Exception as e:
                self.log.exception(error_message(e))
        return converted

class Task(AbstractServerTask):
    def __init__(self, app, name, caption):
        AbstractServerTask.__init__(self, app, name, caption)
        self.on_created = None
        self.on_login = None
        self.on_logout = None
        self.on_ext_request = None
        self.init_dict = {}
        for key, value in self.__dict__.items():
            self.init_dict[key] = value

    @property
    def timeout(self):
        return consts.TIMEOUT

    def login(self, request, dic=None):
        if dic:
            return self.app.login(request, self, dic)
        else:
            return self.app.login(request, self, request.form)

    def logged_in(self, request):
        return self.app.check_session(request, self)

    def redirect(self, location, code=302, Response=None):
        return redirect(location, code, Response)

    def serve_page(self, file_path, dic=None):
        return self.app.serve_page(file_path, dic)

    def copy_database(self, dbtype, connection, limit = 1000):
        copy_database(self, dbtype, connection, limit)


class AdminTask(AbstractServerTask):
    def __init__(self, app, name, caption, db_type, db_database = ''):
        AbstractServerTask.__init__(self, app, name, caption)
        self.timeout = 43200
        self.db_type = db_type
        self.db_info = DBInfo(database=os.path.join(app.work_dir, db_database))
        self.db = get_database(app, db_type, self.db_info.lib)
        self.create_pool()

class Detail(AbstrDetail, ServerDataset):
    def __init__(self, task, owner, name='', caption='', table_name='', master_field=None):
        AbstrDetail.__init__(self, task, owner, name, caption)
        ServerDataset.__init__(self)
        self.item_type_id = consts.DETAIL_TYPE
        self.master_field = master_field
        self.master = owner

    def init_fields(self):
        self.field_defs = []
        for field_def in self.prototype.field_defs:
            self.field_defs.append(list(field_def))
        for field_def in self.field_defs:
            field = DBField(self, field_def)
            self._fields.append(field)
        self.edit_lock = self.prototype.edit_lock
        self._primary_key = self.prototype._primary_key
        self._deleted_flag = self.prototype._deleted_flag
        self._record_version = self.prototype._record_version
        self._master_id = self.prototype._master_id
        self._master_rec_id = self.prototype._master_rec_id
        self._virtual_table = self.prototype._virtual_table

    def do_internal_post(self):
        return {'success': True, 'id': None, 'message': '', 'detail_ids': None}

    def get_filters(self):
        return self.prototype.filters

    def get_reports_info(self):
        return []



#!/usr/bin/env python
print(f"User Guide: https://jampy-docs-v7.readthedocs.io/")
if __name__ == '__main__':
    from jam.wsgi import create_application
    from jam.wsgi_server import run

    application = create_application(__file__)
    run(application)



import os
import sys

from ..common import consts
from .db import AbstractDB

class MSSqlDB(AbstractDB):
    def __init__(self):
        AbstractDB.__init__(self)
        self.db_type = consts.MSSQL
        self.DDL_ROLLBACK = True
        self.IS_DISTINCT_FROM = 'NOT EXISTS(SELECT %s INTERSECT SELECT %s)'
        self.FIELD_TYPES = {
            consts.INTEGER: 'INT',
            consts.TEXT: 'NVARCHAR',
            consts.FLOAT: 'FLOAT',
            consts.CURRENCY: 'FLOAT',
            consts.DATE: 'DATE',
            consts.DATETIME: 'DATETIME',
            consts.BOOLEAN: 'INT',
            consts.LONGTEXT: 'NVARCHAR(MAX)',
            consts.KEYS: 'NVARCHAR(MAX)',
            consts.FILE: 'NVARCHAR(512)',
            consts.IMAGE: 'NVARCHAR(256)'
        }

    def get_params(self, lib):
        params = self.params
        params['name'] = 'MSSQL'
        params['lib'] = ['pymssql', 'pyodbc']
        if lib == 2:
            params['dsn'] = True
            params['database'] = False
        else:
            params['server'] = True
            params['login'] = True
            params['password'] = True
            params['encoding'] = True
            params['host'] = True
            params['port'] = True
        return params

    def get_fields(self, query, fields, alias):
        sql = ''
        for field in fields:
            if field.master_field:
                pass
            else:
                sql += '%s."%s", ' % (alias, field.db_field_name)
        if query.expanded:
            for field in fields:
                if field.lookup_item:
                    sql += '%s_LOOKUP, ' % field.db_field_name
        sql = sql[:-2]
        return sql

    def get_select(self, query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
        offset = query.offset
        limit = query.limit
        if limit:
            end = ''.join([from_clause, where_clause, group_clause])
            offset += 1
            limit += offset
            flds = self.get_fields(query, fields, 'b')
            result = "SELECT %s FROM (SELECT %s, ROW_NUMBER() OVER (%s) AS RowNum FROM %s) AS b WHERE RowNum >= %s AND RowNum < %s ORDER BY RowNum" % \
                (flds, fields_clause, order_clause, end, offset, limit)
        else:
            end = ''.join([from_clause, where_clause, group_clause, order_clause])
            result = 'SELECT %s FROM %s' % (fields_clause, end)
        return result

    def process_query_params(self, params, cursor):
        result = []
        for p in params:
            if type(p) == tuple:
                value, data_type = p
            else:
                value = p
            result.append(value)
        return tuple(result)

    def convert_like(self, field_name, val, data_type):
        if data_type in [consts.INTEGER, consts.FLOAT, consts.CURRENCY]:
            return 'CAST(CAST(%s AS DECIMAL(20, 10)) AS VARCHAR(20))' % field_name, val
        else:
            return field_name, val

    def set_identity_insert(self, table_name, on):
        if on:
            suffix = 'ON'
        else:
            suffix = 'OFF'
        return 'SET IDENTITY_INSERT %s %s' % (table_name, suffix)

    def create_table(self, table_name, fields, gen_name=None, foreign_fields=None):
        result = []
        sql = 'CREATE TABLE "%s"\n(\n' % table_name
        lines = []
        primary_key = None
        for field in fields:
            default_text = self.default_text(field)
            line = '"%s" %s' % (field.field_name, self.FIELD_TYPES[field.data_type])
            if field.size != 0 and field.data_type == consts.TEXT:
                line += '(%d)' % field.size
            if not default_text is None:
##                line += ' CONSTRAINT "%s" DEFAULT %s' % (field.field_name + '_DEFAULT_CONSTRAINT', default_text)          ##fsu 2024.10.26
                line += ' CONSTRAINT "%s" DEFAULT %s' % ('DF_' + table_name + '_' + field.field_name, default_text)         ##fsu 2024.10.26
            if field.primary_key:
                line += ' IDENTITY(1, 1)'
                primary_key = field.field_name
            lines.append(line)
        if primary_key:
            lines.append('CONSTRAINT PK_%s PRIMARY KEY("%s")' % (table_name, primary_key))
        sql += ',\n'.join(lines)
        sql += ')\n'
        return sql

    def drop_table(self, table_name, gen_name):
        return 'DROP TABLE "%s"' % table_name

    def add_field(self, table_name, field):
        default_text = self.default_text(field)
        line = 'ALTER TABLE "%s" ADD "%s" %s' % \
                 (table_name, field.field_name, self.FIELD_TYPES[field.data_type])
        if field.size:
            line += '(%d)' % field.size
        if not default_text is None:
##            line += ' CONSTRAINT "%s" DEFAULT %s' % (field.field_name + '_DEFAULT_CONSTRAINT', default_value)                                                 ##fsu 2024.10.26
            line += ' CONSTRAINT "%s" DEFAULT %s FOR "%s"' % ('DF_' + table_name + '_' + field.field_name, default_text, field.field_name)                  ##fsu 2024.10.26
        print(line)
        return line

    def del_field(self, table_name, field):
        result = 'IF NOT (SELECT COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = \'%s\' AND COLUMN_NAME = \'%s\') IS NULL BEGIN ALTER TABLE "%s" DROP CONSTRAINT "%s" END ALTER TABLE "%s" DROP COLUMN "%s"' % \
            (table_name, field.field_name, table_name, 'DF_' + table_name + '_' + field.field_name, table_name, field.field_name)    ##fsu 2024.10.26
##        return 'ALTER TABLE "%s" DROP COLUMN "%s"' % (table_name, field.field_name)                                                ##fsu 2024.10.26
##        print(result)                                                                                                              ##fsu 2024.10.26
        return result

    def change_field(self, table_name, old_field, new_field):
        if old_field.field_name != new_field.field_name:
            raise Exception("Changing field name is prohibited: field %s" % old_field.field_name)
        result = ''                                                                                                     ##fsu 2024.10.26
 ##        default_value = self.default_value(new_field)                                                                ##fsu 2024.10.26
        default_text = self.default_text(new_field)
        old_default_text = self.default_text(old_field)
        field_info = self.get_field_info(old_field.field_name, table_name)
        if old_field.size != new_field.size:
##            line = 'ALTER TABLE "%s" ALTER COLUMN "%s" %s' % \                                                        ##fsu 2024.10.26
            result += 'ALTER TABLE "%s" ALTER COLUMN "%s" %s' % \
                 (table_name, new_field.field_name, field_info['data_type'])                                            ##fsu 2024.10.26
            size = field_info['size']
            if size and size > 0 and field_info['data_type'].upper() in ['CHAR', 'VARCHAR', 'NVARCHAR']:
                if new_field.size > size:
                    size = new_field.size
##                line += '(%d)' % size                                                                                 ##fsu 2024.10.26
                result += '(%d)' % size
##            result.append(line)                                                                                       ##fsu 2024.10.26
        if old_field.default_value != new_field.default_value:
            if not old_default_text is None:
##                line = 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % \
##                    (table_name, field.field_name + '_DEFAULT_CONSTRAINT')                                            ##fsu 2024.10.26
                result += ' ALTER TABLE "%s" DROP CONSTRAINT "%s"' % \
                    (table_name, 'DF_' + table_name + '_' + old_field.field_name)                                       ##fsu 2024.10.26
            if not default_text is None:
##                line = 'ALTER TABLE "%s" ADD CONSTRAINT "%s" DEFAULT %s' % \
##                    (table_name, field.field_name + '_DEFAULT_CONSTRAINT', default_text)                              ##fsu 2024.10.26
                result += ' ALTER TABLE "%s" ADD CONSTRAINT "%s" DEFAULT %s FOR "%s"' % \
                    (table_name, 'DF_' + table_name + '_' + new_field.field_name, default_text, new_field.field_name)   ##fsu 2024.10.26
##        print(result)                                                                                                 ##fsu 2024.10.26
        return result

    def create_index(self, index_name, table_name, unique, fields, desc):
        return 'CREATE %s INDEX "%s" ON "%s" (%s)' % (unique, index_name, table_name, fields)

    def create_foreign_index(self, table_name, index_name, key, ref, primary_key):
        return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s")' % \
               (table_name, index_name, key, ref, primary_key)

    def drop_index(self, table_name, index_name):
        return 'DROP INDEX "%s" ON "%s"' % (index_name, table_name)

    def drop_foreign_index(self, table_name, index_name):
        return 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % (table_name, index_name)

    def before_insert(self, cursor, pk_field):
        if pk_field and pk_field.data:
            cursor.execute('SET IDENTITY_INSERT %s ON' % pk_field.owner.table_name)

    def after_insert(self, cursor, pk_field):
        if pk_field:
            if not pk_field.data:
                pk_field.data = self.get_lastrowid(cursor)
            else:
                cursor.execute('SET IDENTITY_INSERT %s OFF' % pk_field.owner.table_name)

    def get_table_names(self, connection):
        sql = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = SCHEMA_NAME()"
        try:
            cursor = connection.cursor()
            cursor.execute(sql)
            res = cursor.fetchall()
        except:
            pass
        result = []
        for r in res:
            result.append(r[0])
        return result

    def get_table_info(self, connection, table_name, db_name):
        cursor = connection.cursor()
        sql = '''
            SELECT
                COLUMN_NAME,
                DATA_TYPE,
                CHARACTER_MAXIMUM_LENGTH,
                COLUMN_DEFAULT,
                COLUMNPROPERTY(object_id(TABLE_SCHEMA+'.'+TABLE_NAME), COLUMN_NAME, 'IsIdentity')
            FROM
                INFORMATION_SCHEMA.COLUMNS
            WHERE
              TABLE_NAME = '%s'
        '''
        cursor.execute(sql % (table_name))
        result = cursor.fetchall()
        fields = []
        for column_name, data_type, character_maximum_length, column_default, itent in result:
            size = 0
            if character_maximum_length:
                size = character_maximum_length
            default_value = None
            if column_default:
                default_value = column_default
            pk = False
            if itent == 1:
                pk = True
            fields.append({
                'field_name': column_name,
                'data_type': data_type.upper(),
                'size': size,
                'default_value': default_value,
                'pk': pk
            })
        return {'fields': fields, 'field_types': self.FIELD_TYPES}

db = MSSqlDB()



import os
import json
import sqlite3

from ..common import consts, error_message, file_read, file_write
from ..items import AdminTask, Group
from jam.db.databases import get_database
import jam.langs as langs

# task in this module is admin (Application builder) task

class FieldInfo(object):
    def __init__(self, field, item):
        self.id = field.id.value
        self.field_name = field.f_db_field_name.value
        self.data_type = field.f_data_type.value
        self.size = field.f_size.value
        self.default_value = field.f_default_value.data
        self.master_field = field.f_master_field.value
        self.calc_item = field.f_calc_item.value
        self.primary_key = field.id.value == item.f_primary_key.value

def create_items(task):
    info = file_read(os.path.join(task.app.jam_dir, 'admin', 'builder_structure.info'))
    info = json.loads(info)
    task.set_info(info)

def read_secret_key(task):
    result = None
    con = task.connect()
    try:
        cursor = con.cursor()
        cursor.execute('SELECT F_SECRET_KEY FROM SYS_PARAMS')
        rec = cursor.fetchall()
        result = rec[0][0]
    except:
        pass
    finally:
        con.close()
    if result is None:
        result = ''
    return result

def check_version(task):
    pass

def init_admin(task):
    check_version(task)
    langs.update_langs(task)
    create_items(task)
    update_admin_db(task)
    consts.read_settings()
    consts.MAINTENANCE = False
    consts.write_settings(['MAINTENANCE'])
    consts.read_language()
    from .builder import on_created
    on_created(task)
    from .upgrade import version7_upgrade
    version7_upgrade(task)

def create_admin(app):
    if os.path.exists(os.path.join(app.work_dir, '_admin.sqlite')):
        os.rename(os.path.join(app.work_dir, '_admin.sqlite'), \
            os.path.join(app.work_dir, 'admin.sqlite'))
    task = AdminTask(app, 'admin', 'Administrator', consts.SQLITE,
        db_database=os.path.join(app.work_dir, 'admin.sqlite'))
    app.admin = task
    task.secret_key = read_secret_key(task)
    init_admin(task)
    return task

def connect_task_db(task):
    return task.task_db_module.connect(task.task_db_info)

def valid_delta_type(delta):
    return not delta.f_virtual_table.value and \
        delta.type_id.value in (consts.ITEM_TYPE, consts.TABLE_TYPE)

def get_item_fields(item, fields, delta_fields=None):

    def field_info(field):
        if not field.f_master_field.value and not field.f_calc_item.value:
            return FieldInfo(field, item)

    def fields_info(fields):
        result = []
        for field in fields:
            info = field_info(field)
            if info:
                result.append(info)
        return result

    def find_field(fields_info, field_id):
        for field in fields_info:
            if field.id == field_id:
                return field

    task = item.task
    result = []
    result = fields_info(fields)
    if delta_fields:
        for field in delta_fields:
            if not field.f_master_field.value and not field.f_calc_item.value:
                if field.change_log.record_status == consts.RECORD_INSERTED:
                    info = field_info(field)
                    if info:
                        result.append(info)
                if field.change_log.record_status == consts.RECORD_DELETED:
                    info = find_field(result, field.id.value)
                    if info:
                        result.remove(info)
                elif field.change_log.record_status == consts.RECORD_MODIFIED:
                    info = find_field(result, field.id.value)
                    if info:
                        info.id = field.id.value
                        info.field_name = field.f_db_field_name.value
                        info.data_type = field.f_data_type.value
                        info.size = field.f_size.value
                        info.default_value = field.f_default_value.data
                    else:
                        info = field_info(field)
                        if info:
                            result.append(info)
            elif field.change_log.record_status == consts.RECORD_MODIFIED:
                info = find_field(result, field.id.value)
                if info and not info.master_field and not info.calc_item:
                    result.remove(info)
    return result

def recreate_table(delta, old_fields, new_fields, comp=None, fk_delta=None):

    def foreign_key_dict(ind):
        fields = ind.task.sys_fields.copy()
        fields.set_where(id=ind.f_foreign_field.value)
        fields.open()
        dic = {}
        dic['key'] = fields.f_db_field_name.value
        ref_id = fields.f_object.value
        items = delta.task.sys_items.copy()
        items.set_where(id=ref_id)
        items.open()
        dic['ref'] = items.f_table_name.value
        primary_key = items.f_primary_key.value
        fields.set_where(id=primary_key)
        fields.open()
        dic['primary_key'] = fields.f_db_field_name.value
        return dic

    def get_foreign_fields():
        indices = delta.task.sys_indices.copy()
        indices.set_where(owner_rec_id=delta.id.value)
        indices.open()
        del_id = None
        if fk_delta and (fk_delta.rec_modified() or fk_delta.rec_deleted()):
            del_id = fk_delta.id.value
        result = []
        for ind in indices:
            if ind.f_foreign_index.value:
                if not del_id or ind.id.value != del_id:
                    result.append(foreign_key_dict(ind))
        if fk_delta and (fk_delta.rec_inserted() or fk_delta.rec_modified()):
            result.append(foreign_key_dict(fk_delta))
        return result

    def create_indices_sql():
        indices = delta.task.sys_indices.copy()
        indices.set_where(owner_rec_id=delta.id.value)
        indices.open()
        result = []
        for ind in indices:
            if not ind.f_foreign_index.value:
                result.append(indices_insert_query(ind, delta.f_table_name.value, new_fields=new_fields))
        return result

    def find_field(fields, id_value):
        for f in fields:
            if f.id == id_value:
                return True

    def prepare_fields():
        for f in list(new_fields):
            if not find_field(old_fields, f.id):
                new_fields.remove(f)
        for f in list(old_fields):
            if not find_field(new_fields, f.id):
                old_fields.remove(f)

    connection = connect_task_db(delta.task)
    cursor = connection.cursor()
    db = delta.task.task_db_module
    table_name = delta.f_table_name.value
    result = []
    cursor.execute('PRAGMA foreign_keys=off')
    print('ALTER TABLE "%s" RENAME TO Temp' % table_name)
    cursor.execute('ALTER TABLE "%s" RENAME TO Temp' % table_name)
    try:
        foreign_fields = get_foreign_fields()
        sql = db.create_table(table_name, new_fields, foreign_fields=foreign_fields)
        cursor.execute(sql)
        prepare_fields()
        old_field_list = ['"%s"' % field.field_name for field in old_fields]
        print (old_field_list)
        new_field_list = ['"%s"' % field.field_name for field in new_fields]
        print (new_field_list)
        print ('INSERT INTO "%s" (%s) SELECT %s FROM Temp' % \
            (table_name, ', '.join(new_field_list), ', '.join(old_field_list)))
        cursor.execute('INSERT INTO "%s" (%s) SELECT %s FROM Temp' % \
            (table_name, ', '.join(new_field_list), ', '.join(old_field_list)))
    except Exception as e:
        delta.log.exception(error_message(e))
        cursor.execute('DROP TABLE IF EXISTS "%s"' % table_name)
        cursor.execute('ALTER TABLE Temp RENAME TO "%s"' % table_name)
        cursor.execute('PRAGMA foreign_keys=on')
        connection.commit()
        raise
    cursor.execute('DROP TABLE Temp')
    cursor.execute('PRAGMA foreign_keys=on')
    ind_sql = create_indices_sql()
    for sql in ind_sql:
        cursor.execute(sql)
    connection.commit()
    return result

def change_item_query(delta, old_fields, new_fields):

    def recreate(comp):
        for key, (old_field, new_field) in comp.items():
            if old_field and new_field:
                if old_field.field_name != new_field.field_name:
                    return True
                elif db.default_text(old_field) != db.default_text(new_field):
                    return True
            elif old_field and not new_field:
                return True

    db = delta.task.task_db_module
    table_name = delta.f_table_name.value
    result = []
    comp = {}
    for field in old_fields:
        comp[field.id] = [field, None]
    for field in new_fields:
        if comp.get(field.id):
            comp[field.id][1] = field
        else:
            if field.id:
                comp[field.id] = [None, field]
            else:
                comp[field.field_name] = [None, field]
    if db.db_type == consts.SQLITE and recreate(comp):
        recreate_table(delta, old_fields, new_fields, comp=comp)
        return
    else:
        for key, (old_field, new_field) in comp.items():
            if old_field and not new_field:
                result.append(db.del_field(table_name, old_field))
        for key, (old_field, new_field) in comp.items():
            if old_field and new_field:
                if (old_field.field_name != new_field.field_name) or \
                    (db.FIELD_TYPES[old_field.data_type] != db.FIELD_TYPES[new_field.data_type]) or \
                    (old_field.default_value != new_field.default_value) or \
                    (old_field.size != new_field.size):
                    sql = db.change_field(table_name, old_field, new_field)
                    if type(sql) in (list, tuple):
                        result += sql
                    else:
                        result.append(sql)
        for key, (old_field, new_field) in comp.items():
            if not old_field and new_field:
                result.append(db.add_field(table_name, new_field))
    return result

def insert_item_query(delta, manual_update=False, new_fields=None, foreign_fields=None):
    if not manual_update and valid_delta_type(delta):
        fields = new_fields
        if not fields:
            fields = get_item_fields(delta, delta.sys_fields)
        return delta.task.task_db_module.create_table(delta.f_table_name.value, \
        fields, delta.f_gen_name.value, foreign_fields)

def update_item_query(delta, manual_update=None):
    if not manual_update and valid_delta_type(delta) and delta.sys_fields.rec_count:
        item = delta.copy()
        item.set_where(id=delta.id.value)
        item.open()
        item.sys_fields.open()
        old_fields = get_item_fields(delta, item.sys_fields)
        new_fields = get_item_fields(delta, item.sys_fields, delta.sys_fields)
        return change_item_query(delta, old_fields, new_fields)

def delete_item_query(delta, manual_update=None):
    if not manual_update and valid_delta_type(delta):
        gen_name = None
        if delta.f_primary_key.value:
            gen_name = delta.f_gen_name.value
        return delta.task.task_db_module.drop_table(delta.f_table_name.value, gen_name)

def valid_indices_record(delta):
    items = delta.task.sys_items.copy()
    items.set_where(id=delta.owner_rec_id.value)
    items.open()
    if items.rec_count:
        return not items.f_virtual_table.value
    else:
        return True

def change_foreign_index(delta):
    items = delta.task.sys_items.copy()
    items.set_where(id=delta.owner_rec_id.value)
    items.open()
    it_fields = items.sys_fields
    it_fields.open()
    fields = get_item_fields(items, it_fields)
    new_fields = list(fields)
    recreate_table(items, fields, new_fields, fk_delta=delta)

def create_index(delta, table_name, fields=None, new_fields=None, foreign_key_dict=None):

    def new_field_name_by_id(id_value):
        for f in new_fields:
            if f.id == id_value:
                return f.field_name

    db = delta.task.task_db_module
    index_name = delta.f_index_name.value
    if delta.f_foreign_index.value:
        if foreign_key_dict:
            key = foreign_key_dict['key']
            ref = foreign_key_dict['ref']
            primary_key = foreign_key_dict['primary_key']
        else:
            fields = delta.task.sys_fields.copy()
            fields.set_where(id=delta.f_foreign_field.value)
            fields.open()
            key = fields.f_db_field_name.value
            ref_id = fields.f_object.value
            items = delta.task.sys_items.copy()
            items.set_where(id=ref_id)
            items.open()
            ref = items.f_table_name.value
            primary_key = items.f_primary_key.value
            fields.set_where(id=primary_key)
            fields.open()
            primary_key = fields.f_db_field_name.value
        sql = db.create_foreign_index(table_name, index_name, key, ref, primary_key)
    else:
        index_fields = delta.f_fields_list.value
        desc = ''
        if delta.descending.value:
            desc = 'DESC'
        unique = ''
        if delta.f_unique_index.value:
            unique = 'UNIQUE'
        fields = delta.load_index_fields(index_fields)
        if db.db_type == consts.FIREBIRD:
            if new_fields:
                field_defs = [new_field_name_by_id(field[0]) for field in fields]
            else:
                field_defs = [delta.task.sys_fields.field_by_id(field[0], 'f_db_field_name') for field in fields]
            field_str = '"' + '", "'.join(field_defs) + '"'
        else:
            field_defs = []
            for field in fields:
                if new_fields:
                    field_name = new_field_name_by_id(field[0])
                else:
                    field_name = delta.task.sys_fields.field_by_id(field[0], 'f_db_field_name')
                d = ''
                if field[1]:
                    d = 'DESC'
                field_defs.append('"%s" %s' % (field_name, d))
            field_str = ', '.join(field_defs)
        sql = db.create_index(index_name, table_name, unique, field_str, desc)
    return sql

def indices_insert_query(delta, table_name=None, new_fields=None, manual_update=False, foreign_key_dict=None):
    if not manual_update and valid_indices_record(delta):
        if not table_name:
            table_name = delta.task.sys_items.field_by_id(delta.owner_rec_id.value, 'f_table_name')
        if delta.task.task_db_module.db_type == consts.SQLITE and delta.f_foreign_index.value:
            if not new_fields:
                change_foreign_index(delta)
        else:
            return create_index(delta, table_name, new_fields=new_fields, foreign_key_dict=foreign_key_dict)

def indices_delete_query(delta, table_name=None, manual_update=False):
    if not manual_update and valid_indices_record(delta):
        if delta.task.task_db_module.db_type == consts.SQLITE and delta.f_foreign_index.value:
            change_foreign_index(delta)
        else:
            db = delta.task.task_db_module
            if not table_name:
                table_name = delta.task.sys_items.field_by_id(delta.owner_rec_id.value, 'f_table_name')
            index_name = delta.f_index_name.value
            if delta.f_foreign_index.value:
                return db.drop_foreign_index(table_name, index_name)
            else:
                return db.drop_index(table_name, index_name)

def update_admin_db(task):

    def do_updates(con, field, item_name):
        if item_name == 'sys_privileges' and field.field_name.lower() == 'owner_item':
            cursor = con.cursor()
            cursor.execute("SELECT ID FROM SYS_ITEMS WHERE TABLE_ID > 0 AND DELETED = 0")
            details = cursor.fetchall()
            cursor.execute("SELECT ID FROM SYS_ROLES WHERE DELETED = 0")
            roles = cursor.fetchall()
            for d in details:
                for r in roles:
                    cursor.execute("""
                        INSERT INTO SYS_PRIVILEGES
                        (DELETED, OWNER_ID, OWNER_REC_ID, ITEM_ID, F_CAN_VIEW, F_CAN_CREATE, F_CAN_EDIT, F_CAN_DELETE)
                        values (?, ?, ?, ?, ?, ?, ?, ?)""",
                        (0, 2, r[0], d[0], True, True, True, True))
            con.commit()

    def get_item_fields(item, table_name):
        cursor.execute('PRAGMA table_info(%s)' % table_name)
        rows = cursor.fetchall()
        result = [str(row[1]).upper() for row in rows]
        return result

    def check_item_fields(item, table_name=None):
        if not table_name:
            table_name = item.table_name.upper()
        fields = get_item_fields(item, table_name)
        for field in item._fields:
            if not field.field_name.upper() in fields:
                sql = 'ALTER TABLE %s ADD COLUMN %s %s' % \
                    (table_name, field.field_name.upper(), \
                    task.db.FIELD_TYPES[field.data_type])
                cursor.execute(sql)
                con.commit()
                do_updates(con, field, item.item_name)

    def check_table_exists(item, table_name=None):
        if not table_name:
            table_name = item.table_name.upper()
        sql = 'SELECT name FROM sqlite_master WHERE type="table" AND UPPER(name)="%s"' % table_name
        cursor.execute(sql)
        rows = cursor.fetchall()
        if not rows:
            sql = 'CREATE TABLE %s (ID INTEGER PRIMARY KEY)' % table_name
            cursor.execute(sql)
        return True

    con = task.connect()
    try:
        cursor = con.cursor()
        for group in task.items:
            for item in group.items:
                if item.table_name and not item.master:
                    if check_table_exists(item):
                        check_item_fields(item)
    finally:
        con.close()

def get_privileges(task):
    priv = {}
    r = task.sys_roles.copy()
    r.open()
    privliges = task.sys_privileges.copy()
    privliges.open()
    for r in r:
        priv[r.id.value] = {}
    for p in privliges:
        priv[p.owner_rec_id.value][p.item_id.data] = \
            {
                'can_view': p.f_can_view.value,
                'can_create': p.f_can_create.value,
                'can_edit': p.f_can_edit.value,
                'can_delete': p.f_can_delete.value
            }
    return priv

def get_field_restrictions(task):
    result = {}
    r = task.sys_roles.copy()
    r.open()
    field_restrictions = task.sys_field_privileges.copy()
    field_restrictions.open()
    for r in r:
        result[r.id.value] = {}
    for p in field_restrictions:
        result[p.owner_rec_id.value][p.field.data] = \
            {
                'prohibited': p.f_prohibited.value,
                'read_only': p.f_read_only.value
            }
    return result

def login_user(task, log, password, admin, ip=None, session_uuid=None):
    user_id = None
    user_info = {}
    if consts.SAFE_MODE:
        users = task.sys_users.copy()
        users.set_where(f_password=password)
        users.open()
        for u in users:
            if u.f_login.value.strip() == log.strip() and u.f_password.value == password:
                if not admin or u.f_admin.value == admin:
                    user_id = u.id.value
                    user_info = {
                        'user_id': u.id.value,
                        'role_id': u.f_role.value,
                        'role_name': u.f_role.display_text,
                        'user_name': u.f_name.value,
                        'admin': u.f_admin.value
                    }
                    if ip or session_uuid:
                        task.execute("UPDATE SYS_USERS SET F_IP='%s', F_UUID='%s' WHERE ID=%s" % (ip, session_uuid, u.id.value))
                    break
    return user_info

def user_valid_ip(task, user_id, ip):
    res = task.select("SELECT F_IP FROM SYS_USERS WHERE ID=%s" % user_id)
    if res and res[0][0] == ip:
        return True
    return False

def user_valid_uuid(task, user_id, session_uuid):
    res = task.select("SELECT F_UUID FROM SYS_USERS WHERE ID=%s" % user_id)
    if res and res[0][0] == session_uuid:
        return True
    return False

def indexes_get_table_names(indexes):
    ids = []
    for i in indexes:
        ids.append(i.owner_rec_id.value)
    items = indexes.task.sys_items.copy(handlers=False)
    items.set_where(id__in=ids)
    items.open(fields=['id', 'f_table_name'])
    table_names = {}
    for i in items:
        table_names[i.id.value] = i.f_table_name.value
    return table_names

def drop_indexes_sql(task):
    db = task.task_db_module
    indexes = task.sys_indices.copy(handlers=False)
    indexes.open()
    table_names = indexes_get_table_names(indexes)
    sqls = []
    for i in indexes:
        if not (i.f_foreign_index.value and db.db_type == consts.SQLITE):
            table_name = table_names.get(i.owner_rec_id.value)
            if table_name:
                sqls.append(indices_delete_query(i, table_name))
    return sqls

def restore_indexes_sql(task):
    db = task.task_db_module
    indexes = task.sys_indices.copy(handlers=False)
    indexes.open()
    table_names = indexes_get_table_names(indexes)
    sqls = []
    for i in indexes:
        if not (i.f_foreign_index.value and db.db_type == consts.SQLITE):
            table_name = table_names.get(i.owner_rec_id.value)
            if table_name:
                sqls.append(indices_insert_query(i, table_name))
    return sqls
